from __future__ import annotations

import threading
import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed

import numpy as np
from numpy.linalg import norm
from openai import OpenAI
from PIL import Image
from pydantic.fields import FieldInfo
from sentence_transformers import SentenceTransformer

from palimpzest.constants import (
    MODEL_CARDS,
    NAIVE_EST_JOIN_SELECTIVITY,
    NAIVE_EST_NUM_INPUT_TOKENS,
    Cardinality,
    Model,
    PromptStrategy,
)
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.lib.schemas import AUDIO_FIELD_TYPES, IMAGE_FIELD_TYPES, ImageFilepath
from palimpzest.core.models import GenerationStats, OperatorCostEstimates, RecordOpStats
from palimpzest.query.generators.generators import Generator
from palimpzest.query.operators.physical import PhysicalOperator


class Singleton:
     def __new__(cls, *args, **kw):
         if not hasattr(cls, '_instance'):
             orig = super(Singleton, cls)  # noqa: UP008
             cls._instance = orig.__new__(cls, *args, **kw)
         return cls._instance

class Locks(Singleton):
    model = None
    clip_lock = threading.Lock()
    exec_lock = threading.Lock()

    @classmethod
    def get_model(cls, model_name: str):
        with cls.clip_lock:
            if cls.model is None:
                cls.model = SentenceTransformer(model_name)
            return cls.model

def compute_similarity(left_embedding: list[float], right_embedding: list[float]) -> float:
    """
    Compute the similarity between two embeddings using cosine similarity.
    """
    return np.dot(left_embedding, right_embedding) / (norm(left_embedding) * norm(right_embedding))


class JoinOp(PhysicalOperator, ABC):
    def __init__(
        self,
        condition: str,
        how: str = "inner",
        on: list[str] | None = None,
        join_parallelism: int = 64,
        retain_inputs: bool = True,
        desc: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        assert self.input_schema == self.output_schema, "Input and output schemas must match for JoinOp"
        self.condition = condition
        self.how = how
        self.on = on
        self.join_parallelism = join_parallelism
        self.retain_inputs = retain_inputs
        self.desc = desc
        self.join_idx = 0
        self.finished = False

        # maintain list(s) of input records for the join
        self._left_input_records: list[DataRecord] = []
        self._right_input_records: list[DataRecord] = []

        # maintain set of left/right record ids that have been joined (for left/right/outer joins)
        self._left_joined_record_ids: set[str] = set()
        self._right_joined_record_ids: set[str] = set()

    def __str__(self):
        op = super().__str__()
        op += f"    Condition: {self.condition}\n"
        op += f"    How: {self.how}\n"
        op += f"    On: {self.on}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {
            "condition": self.condition,
            "join_parallelism": self.join_parallelism,
            "desc": self.desc,
            "how": self.how,
            "on": self.on,
            **id_params,
        }
        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {
            "condition": self.condition,
            "join_parallelism": self.join_parallelism,
            "retain_inputs": self.retain_inputs,
            "desc": self.desc,
            "how": self.how,
            "on": self.on,
            **op_params,
        }
        return op_params

    def _compute_unmatched_records(self) -> DataRecordSet:
        """Helper function to compute unmatched records for left/right/outer joins."""
        def join_unmatched_records(input_records: list[DataRecord] | list[tuple[DataRecord, list[float]]], joined_record_ids: set[str], left: bool = True):
            records, record_op_stats_lst = [], []
            for record in input_records:
                start_time = time.time()
                record = record[0] if isinstance(record, tuple) else record
                if record._id not in joined_record_ids:
                    unmatched_dr = (
                        DataRecord.from_join_parents(self.output_schema, record, None)
                        if left
                        else DataRecord.from_join_parents(self.output_schema, None, record)
                    )
                    unmatched_dr._passed_operator = True

                    # compute record stats and add to output_record_op_stats
                    time_per_record = time.time() - start_time
                    record_op_stats = RecordOpStats(
                        record_id=unmatched_dr._id,
                        record_parent_ids=unmatched_dr._parent_ids,
                        record_source_indices=unmatched_dr._source_indices,
                        record_state=unmatched_dr.to_dict(include_bytes=False),
                        full_op_id=self.get_full_op_id(),
                        logical_op_id=self.logical_op_id,
                        op_name=self.op_name(),
                        time_per_record=time_per_record,
                        cost_per_record=0.0,
                        model_name=self.get_model_name(),
                        join_condition=str(self.on),
                        fn_call_duration_secs=time_per_record,
                        answer={"passed_operator": True},
                        passed_operator=True,
                        op_details={k: str(v) for k, v in self.get_id_params().items()},
                    )
                    records.append(unmatched_dr)
                    record_op_stats_lst.append(record_op_stats)
            return records, record_op_stats_lst

        records, record_op_stats = [], []
        if self.how == "left":
            records, record_op_stats = join_unmatched_records(self._left_input_records, self._left_joined_record_ids, left=True)

        elif self.how == "right":
            records, record_op_stats = join_unmatched_records(self._right_input_records, self._right_joined_record_ids, left=False)

        elif self.how == "outer":
            records, record_op_stats = join_unmatched_records(self._left_input_records, self._left_joined_record_ids, left=True)
            right_records, right_record_op_stats = join_unmatched_records(self._right_input_records, self._right_joined_record_ids, left=False)
            records.extend(right_records)
            record_op_stats.extend(right_record_op_stats)

        return DataRecordSet(records, record_op_stats)

    @abstractmethod
    def naive_cost_estimates(self, left_source_op_cost_estimates: OperatorCostEstimates, right_source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        pass

    def set_finished(self):
        """Mark the operator as finished after computing left/right/outer join logic."""
        self.finished = True

class RelationalJoin(JoinOp):

    def get_model_name(self):
        return None
    
    def _process_join_candidate_pair(self, left_candidate, right_candidate) -> tuple[DataRecord, RecordOpStats]:
        start_time = time.time()

        # determine whether or not the join was satisfied
        passed_operator = all(
            left_candidate[field] == right_candidate[field]
            for field in self.on
        )

        # handle different join types
        if self.how == "left" and passed_operator:
            self._left_joined_record_ids.add(left_candidate._id)
        elif self.how == "right" and passed_operator:
            self._right_joined_record_ids.add(right_candidate._id)
        elif self.how == "outer" and passed_operator:
            self._left_joined_record_ids.add(left_candidate._id)
            self._right_joined_record_ids.add(right_candidate._id)

        # compute output record and add to output_records
        join_dr = DataRecord.from_join_parents(self.output_schema, left_candidate, right_candidate)
        join_dr._passed_operator = passed_operator

        # compute record stats and add to output_record_op_stats
        time_per_record = time.time() - start_time
        record_op_stats = RecordOpStats(
            record_id=join_dr._id,
            record_parent_ids=join_dr._parent_ids,
            record_source_indices=join_dr._source_indices,
            record_state=join_dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=time_per_record,
            cost_per_record=0.0,
            model_name=self.get_model_name(),
            join_condition=str(self.on),
            fn_call_duration_secs=time_per_record,
            answer={"passed_operator": passed_operator},
            passed_operator=passed_operator,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return join_dr, record_op_stats

    def naive_cost_estimates(self, left_source_op_cost_estimates: OperatorCostEstimates, right_source_op_cost_estimates: OperatorCostEstimates):
        # estimate output cardinality using a constant assumption of the filter selectivity
        selectivity = NAIVE_EST_JOIN_SELECTIVITY
        cardinality = selectivity * (left_source_op_cost_estimates.cardinality * right_source_op_cost_estimates.cardinality)

        # estimate 1 ms execution time per input record pair
        time_per_record = 0.001 * (left_source_op_cost_estimates.cardinality + right_source_op_cost_estimates.cardinality)

        return OperatorCostEstimates(
            cardinality=cardinality,
            time_per_record=time_per_record,
            cost_per_record=0.0,
            quality=1.0,
        )

    def __call__(self, left_candidates: list[DataRecord], right_candidates: list[DataRecord], final: bool = False) -> tuple[DataRecordSet, int]:
        # create the set of candidates to join
        join_candidates = []
        for candidate in left_candidates:
            for right_candidate in right_candidates:
                join_candidates.append((candidate, right_candidate))
            for right_candidate in self._right_input_records:
                join_candidates.append((candidate, right_candidate))
        for candidate in self._left_input_records:
            for right_candidate in right_candidates:
                join_candidates.append((candidate, right_candidate))

        # apply the join logic to each pair of candidates
        output_records, output_record_op_stats = [], []
        with ThreadPoolExecutor(max_workers=self.join_parallelism) as executor:
            futures = [
                executor.submit(self._process_join_candidate_pair, candidate, right_candidate)
                for candidate, right_candidate in join_candidates
            ]
  
            # collect results as they complete
            for future in as_completed(futures):
                self.join_idx += 1
                join_output_record, join_output_record_op_stats = future.result()
                output_records.append(join_output_record)
                output_record_op_stats.append(join_output_record_op_stats)

        # compute the number of inputs processed
        num_inputs_processed = len(join_candidates)

        # store input records to join with new records added later
        if self.retain_inputs:
            self._left_input_records.extend(left_candidates)
            self._right_input_records.extend(right_candidates)
        
        # if this is the final call, then add in any left/right/outer join records that did not match
        if final:
            return self._compute_unmatched_records(), 0

        # return empty DataRecordSet if no output records were produced
        if len(output_records) == 0:
            return DataRecordSet([], []), num_inputs_processed

        return DataRecordSet(output_records, output_record_op_stats), num_inputs_processed



class LLMJoin(JoinOp):
    def __init__(
        self,
        model: Model,
        prompt_strategy: PromptStrategy = PromptStrategy.JOIN,
        reasoning_effort: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.model = model
        self.prompt_strategy = prompt_strategy
        self.reasoning_effort = reasoning_effort
        self.generator = Generator(model, prompt_strategy, reasoning_effort, self.api_base, Cardinality.ONE_TO_ONE, self.desc, self.verbose)

    def __str__(self):
        op = super().__str__()
        op += f"    Model: {self.model.value}\n"
        op += f"    Reasoning Effort: {self.reasoning_effort}\n"
        op += f"    Prompt Strategy: {self.prompt_strategy.value}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {
            "model": self.model.value,
            "prompt_strategy": self.prompt_strategy.value,
            "reasoning_effort": self.reasoning_effort,
            **id_params,
        }
        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {
            "model": self.model,
            "prompt_strategy": self.prompt_strategy,
            "reasoning_effort": self.reasoning_effort,
            **op_params,
        }
        return op_params

    def get_model_name(self):
        return self.model.value

    def _process_join_candidate_pair(
        self,
        left_candidate: DataRecord,
        right_candidate: DataRecord,
        gen_kwargs: dict,
    ) -> tuple[DataRecord, RecordOpStats]:
        start_time = time.time()

        # generate output; NOTE: FieldInfo is used to indicate the output type; thus, the desc is not needed
        fields = {"passed_operator": FieldInfo(annotation=bool, description="Whether the records satisfy the join condition")}
        field_answers, _, generation_stats, _ = self.generator(left_candidate, fields, right_candidate=right_candidate, **gen_kwargs)

        # determine whether or not the join was satisfied
        passed_operator = field_answers["passed_operator"]

        # handle different join types
        if self.how == "left" and passed_operator:
            self._left_joined_record_ids.add(left_candidate._id)
        elif self.how == "right" and passed_operator:
            self._right_joined_record_ids.add(right_candidate._id)
        elif self.how == "outer" and passed_operator:
            self._left_joined_record_ids.add(left_candidate._id)
            self._right_joined_record_ids.add(right_candidate._id)

        # compute output record and add to output_records
        join_dr = DataRecord.from_join_parents(self.output_schema, left_candidate, right_candidate)
        join_dr._passed_operator = passed_operator

        # compute record stats and add to output_record_op_stats
        record_op_stats = RecordOpStats(
            record_id=join_dr._id,
            record_parent_ids=join_dr._parent_ids,
            record_source_indices=join_dr._source_indices,
            record_state=join_dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=time.time() - start_time,
            cost_per_record=generation_stats.cost_per_record,
            model_name=self.get_model_name(),
            join_condition=self.condition,
            total_input_tokens=generation_stats.total_input_tokens,
            total_output_tokens=generation_stats.total_output_tokens,
            total_embedding_input_tokens=generation_stats.total_embedding_input_tokens,
            total_input_cost=generation_stats.total_input_cost,
            total_output_cost=generation_stats.total_output_cost,
            total_embedding_cost=generation_stats.total_embedding_cost,
            llm_call_duration_secs=generation_stats.llm_call_duration_secs,
            fn_call_duration_secs=generation_stats.fn_call_duration_secs,
            total_llm_calls=generation_stats.total_llm_calls,
            total_embedding_llm_calls=generation_stats.total_embedding_llm_calls,
            answer=field_answers,
            passed_operator=passed_operator,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return join_dr, record_op_stats


class NestedLoopsJoin(LLMJoin):

    def naive_cost_estimates(self, left_source_op_cost_estimates: OperatorCostEstimates, right_source_op_cost_estimates: OperatorCostEstimates):
        # estimate number of input tokens from source
        est_num_input_tokens = 2 * NAIVE_EST_NUM_INPUT_TOKENS
        if self.is_image_op():
            est_num_input_tokens = 2 * 765 / 10  # 1024x1024 image is 765 tokens

        # NOTE: the output often generates an entire reasoning sentence, thus the true value may be higher
        # the filter operation's LLM call should only output TRUE or FALSE, thus we expect its
        # number of output tokens to be ~1.25
        est_num_output_tokens = 1.25

        # get est. of conversion time per record from model card;
        model_conversion_time_per_record = (
            MODEL_CARDS[self.model.value]["seconds_per_output_token"] * est_num_output_tokens
        )

        # get est. of conversion cost (in USD) per record from model card
        usd_per_input_token = (
            MODEL_CARDS[self.model.value]["usd_per_audio_input_token"]
            if self.is_audio_op()
            else MODEL_CARDS[self.model.value]["usd_per_input_token"]
        )
        model_conversion_usd_per_record = (
            usd_per_input_token * est_num_input_tokens
            + MODEL_CARDS[self.model.value]["usd_per_output_token"] * est_num_output_tokens
        )

        # estimate output cardinality using a constant assumption of the filter selectivity
        selectivity = NAIVE_EST_JOIN_SELECTIVITY
        cardinality = selectivity * (left_source_op_cost_estimates.cardinality * right_source_op_cost_estimates.cardinality)

        # estimate quality of output based on the strength of the model being used
        quality = (MODEL_CARDS[self.model.value]["overall"] / 100.0)

        return OperatorCostEstimates(
            cardinality=cardinality,
            time_per_record=model_conversion_time_per_record,
            cost_per_record=model_conversion_usd_per_record,
            quality=quality,
        )

    def __call__(self, left_candidates: list[DataRecord], right_candidates: list[DataRecord], final: bool = False) -> tuple[DataRecordSet, int]:
        # get the set of input fields from both records in the join
        input_fields = self.get_input_fields()

        # construct kwargs for generation
        gen_kwargs = {"project_cols": input_fields, "join_condition": self.condition}

        # create the set of candidates to join
        join_candidates = []
        for candidate in left_candidates:
            for right_candidate in right_candidates:
                join_candidates.append((candidate, right_candidate))
            for right_candidate in self._right_input_records:
                join_candidates.append((candidate, right_candidate))
        for candidate in self._left_input_records:
            for right_candidate in right_candidates:
                join_candidates.append((candidate, right_candidate))

        # apply the generator to each pair of candidates
        output_records, output_record_op_stats = [], []
        with ThreadPoolExecutor(max_workers=self.join_parallelism) as executor:
            futures = [
                executor.submit(self._process_join_candidate_pair, candidate, right_candidate, gen_kwargs)
                for candidate, right_candidate in join_candidates
            ]
  
            # collect results as they complete
            for future in as_completed(futures):
                self.join_idx += 1
                join_output_record, join_output_record_op_stats = future.result()
                output_records.append(join_output_record)
                output_record_op_stats.append(join_output_record_op_stats)
                print(f"{self.join_idx} JOINED")

        # compute the number of inputs processed
        num_inputs_processed = len(join_candidates)

        # store input records to join with new records added later
        if self.retain_inputs:
            self._left_input_records.extend(left_candidates)
            self._right_input_records.extend(right_candidates)

        # if this is the final call, then add in any left/right/outer join records that did not match
        if final:
            return self._compute_unmatched_records(), 0

        # return empty DataRecordSet if no output records were produced
        if len(output_records) == 0:
            return DataRecordSet([], []), num_inputs_processed

        return DataRecordSet(output_records, output_record_op_stats), num_inputs_processed


class EmbeddingJoin(LLMJoin):
    # NOTE: we currently do not support audio joins as embedding models for audio seem to have
    # specialized use cases (e.g., speech-to-text) with strict requirements on things like e.g. sample rate
    def __init__(
        self,
        num_samples: int = 10,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.num_samples = num_samples
        self.samples_drawn = 0

        # compute whether all fields are text fields
        self.text_only = all([
            field.annotation not in IMAGE_FIELD_TYPES + AUDIO_FIELD_TYPES
            for field_name, field in self.input_schema.model_fields.items()
            if field_name.split(".")[-1] in self.get_input_fields()
        ])
        self.embedding_model = Model.TEXT_EMBEDDING_3_SMALL if self.text_only else Model.CLIP_VIT_B_32
        self.locks = Locks()

        # keep track of embedding costs that could not be amortized if no output records were produced
        self.residual_embedding_cost = 0.0

        # crude adjustment factor for naive estimation in unoptimized setting
        self.naive_quality_adjustment = 0.6

        # maintain list(s) of input records and their embeddings for the join
        self._left_input_records: list[tuple[DataRecord, list[float]]] = []
        self._right_input_records: list[tuple[DataRecord, list[float]]] = []

        # maintain lowest and highest embedding similarities for matching and non-matching pairs
        self.min_matching_sim = float("inf")
        self.max_non_matching_sim = float("-inf")

    def __str__(self):
        op = super().__str__()
        op += f"    Num Samples: {self.num_samples}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {
            "num_samples": self.num_samples,
            **id_params,
        }

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {
            "num_samples": self.num_samples,
            **op_params,
        }

        return op_params

    def naive_cost_estimates(self, left_source_op_cost_estimates: OperatorCostEstimates, right_source_op_cost_estimates: OperatorCostEstimates):
        # estimate number of input tokens from source
        est_num_input_tokens = 2 * NAIVE_EST_NUM_INPUT_TOKENS
        if self.is_image_op():
            est_num_input_tokens = 2 * 765 / 10  # 1024x1024 image is 765 tokens

        # NOTE: the output often generates an entire reasoning sentence, thus the true value may be higher
        # the filter operation's LLM call should only output TRUE or FALSE, thus we expect its
        # number of output tokens to be ~1.25
        est_num_output_tokens = 1.25

        # get est. of conversion time per record from model card;
        model_conversion_time_per_record = (
            MODEL_CARDS[self.embedding_model.value]["seconds_per_output_token"] * est_num_output_tokens
        )

        # get est. of conversion cost (in USD) per record from model card
        model_conversion_usd_per_record = MODEL_CARDS[self.embedding_model.value]["usd_per_input_token"] * est_num_input_tokens

        # estimate output cardinality using a constant assumption of the filter selectivity
        selectivity = NAIVE_EST_JOIN_SELECTIVITY
        cardinality = selectivity * (left_source_op_cost_estimates.cardinality * right_source_op_cost_estimates.cardinality)

        # estimate quality of output based on the strength of the model being used
        quality = (MODEL_CARDS[self.model.value]["overall"] / 100.0) * self.naive_quality_adjustment

        return OperatorCostEstimates(
            cardinality=cardinality,
            time_per_record=model_conversion_time_per_record,
            cost_per_record=model_conversion_usd_per_record,
            quality=quality,
        )

    def _compute_embeddings(self, candidates: list[DataRecord], input_fields: list[str]) -> tuple[np.ndarray, GenerationStats]:
        # return empty array and empty stats if no candidates  
        if len(candidates) == 0:
            return np.zeros((0, 512)), GenerationStats()

        start_time = time.time()
        total_embedding_input_tokens = 0
        embeddings = None
        if self.text_only:
            client = OpenAI()
            inputs = [dr.to_json_str(bytes_to_str=True, project_cols=input_fields, sorted=True) for dr in candidates]
            response = client.embeddings.create(input=inputs, model=self.embedding_model.value)
            total_embedding_input_tokens = response.usage.total_tokens
            embeddings = np.array([item.embedding for item in response.data])
        else:
            model = self.locks.get_model(self.embedding_model.value)
            embeddings = np.zeros((len(candidates), 512))  # CLIP embeddings are 512-dimensional
            num_input_fields_present = 0
            for field in input_fields:
                field_inputs = []
                for candidate in candidates:
                    if field not in candidate.get_field_names():
                        continue
                    num_input_fields_present += 1
                    field_type = candidate.get_field_type(field)
                    if field_type in [ImageFilepath]:
                        field_inputs.append(Image.open(candidate[field]))
                    else:
                        field_inputs.append(str(candidate[field]))
                
                if len(field_inputs) > 0:
                    embeddings += model.encode(field_inputs, convert_to_numpy=True)

            # average embeddings over input fields present in candidates
            embeddings /= num_input_fields_present

        # compute cost of embedding(s)
        model_card = MODEL_CARDS[self.embedding_model.value]
        total_embedding_cost = model_card["usd_per_input_token"] * total_embedding_input_tokens
        embedding_gen_stats = GenerationStats(
            model_name=self.embedding_model.value,
            total_input_tokens=0.0,
            total_output_tokens=0.0,
            total_embedding_input_tokens=total_embedding_input_tokens,
            total_input_cost=0.0,
            total_output_cost=0.0,
            total_embedding_cost=total_embedding_cost,
            cost_per_record=total_embedding_cost,
            llm_call_duration_secs=time.time() - start_time,
            total_llm_calls=1,
            total_embedding_llm_calls=len(candidates),
        )

        return embeddings, embedding_gen_stats

    def _process_join_candidate_pair(self, left_candidate, right_candidate, gen_kwargs, embedding_sim):
        output_record, output_record_op_stats = super()._process_join_candidate_pair(left_candidate, right_candidate, gen_kwargs)
        return output_record, output_record_op_stats, embedding_sim

    def _process_join_candidate_with_sim(self, left_candidate: DataRecord, right_candidate: DataRecord, embedding_sim: float, passed_operator: bool) -> tuple[DataRecord, RecordOpStats]:
        # compute output record and add to output_records
        join_dr = DataRecord.from_join_parents(self.output_schema, left_candidate, right_candidate)
        join_dr._passed_operator = passed_operator

        # handle different join types
        if self.how == "left" and passed_operator:
            self._left_joined_record_ids.add(left_candidate._id)
        elif self.how == "right" and passed_operator:
            self._right_joined_record_ids.add(right_candidate._id)
        elif self.how == "outer" and passed_operator:
            self._left_joined_record_ids.add(left_candidate._id)
            self._right_joined_record_ids.add(right_candidate._id)

        # NOTE: embedding costs are amortized over all records and added at the end of __call__
        # compute record stats and add to output_record_op_stats
        record_op_stats = RecordOpStats(
            record_id=join_dr._id,
            record_parent_ids=join_dr._parent_ids,
            record_source_indices=join_dr._source_indices,
            record_state=join_dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=0.0,
            cost_per_record=0.0,
            model_name=self.get_model_name(),
            join_condition=self.condition,
            answer={"passed_operator": passed_operator},
            passed_operator=passed_operator,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return join_dr, record_op_stats, embedding_sim

    def __call__(self, left_candidates: list[DataRecord], right_candidates: list[DataRecord], final: bool = False) -> tuple[DataRecordSet, int]:
        # get the set of input fields from both records in the join
        input_fields = self.get_input_fields()

        # compute the embeding for each candidate
        left_embeddings, left_embedding_gen_stats = self._compute_embeddings(left_candidates, input_fields)
        right_embeddings, right_embedding_gen_stats = self._compute_embeddings(right_candidates, input_fields)
        total_embedding_cost = left_embedding_gen_stats.cost_per_record + right_embedding_gen_stats.cost_per_record + self.residual_embedding_cost
        self.residual_embedding_cost = 0.0

        # construct kwargs for generation
        gen_kwargs = {"project_cols": input_fields, "join_condition": self.condition}

        # TODO: add embeddings to join candidates
        # create the set of candidates to join
        join_candidates = []
        for candidate, embedding in zip(left_candidates, left_embeddings):
            for right_candidate, right_embedding in zip(right_candidates, right_embeddings):
                embedding_sim = compute_similarity(embedding, right_embedding)
                join_candidates.append((candidate, right_candidate, embedding_sim))
            for right_candidate, right_embedding in self._right_input_records:
                embedding_sim = compute_similarity(embedding, right_embedding)
                join_candidates.append((candidate, right_candidate, embedding_sim))
        for candidate, embedding in self._left_input_records:
            for right_candidate, right_embedding in zip(right_candidates, right_embeddings):
                embedding_sim = compute_similarity(embedding, right_embedding)
                join_candidates.append((candidate, right_candidate, embedding_sim))

        # prepare list of output records and their stats
        output_records, output_record_op_stats, num_inputs_processed = [], [], 0

        # draw samples until num_samples is reached
        with self.locks.exec_lock:
            if self.samples_drawn < self.num_samples:
                samples_to_draw = min(self.num_samples - self.samples_drawn, len(join_candidates))
                join_candidate_samples = join_candidates[:samples_to_draw]
                join_candidates = join_candidates[samples_to_draw:]

                # apply the generator to each pair of candidates
                with ThreadPoolExecutor(max_workers=self.join_parallelism) as executor:
                    futures = [
                        executor.submit(self._process_join_candidate_pair, left_candidate, right_candidate, gen_kwargs, embedding_sim)
                        for left_candidate, right_candidate, embedding_sim in join_candidate_samples
                    ]

                    # collect results as they complete
                    similarities, joined = [], []
                    for future in as_completed(futures):
                        self.join_idx += 1
                        join_output_record, join_output_record_op_stats, embedding_sim = future.result()
                        output_records.append(join_output_record)
                        output_record_op_stats.append(join_output_record_op_stats)
                        similarities.append(embedding_sim)
                        joined.append(join_output_record._passed_operator)
                        print(f"{self.join_idx} JOINED")

                    # sort join results by embedding similarity
                    sorted_sim_join_tuples = sorted(zip(similarities, joined), key=lambda x: x[0])

                    # compute threshold below which no records joined
                    for embedding_sim, records_joined in sorted_sim_join_tuples:
                        if records_joined:
                            break
                        if not records_joined and embedding_sim > self.max_non_matching_sim:
                            self.max_non_matching_sim = embedding_sim

                    # compute threshold above which all records joined
                    for embedding_sim, records_joined in reversed(sorted_sim_join_tuples):
                        if not records_joined:
                            break
                        if records_joined and embedding_sim < self.min_matching_sim:
                            self.min_matching_sim = embedding_sim

                # update samples drawn and num_inputs_processed
                self.samples_drawn += samples_to_draw
                num_inputs_processed += samples_to_draw

        # process remaining candidates based on embedding similarity
        if len(join_candidates) > 0:
             assert self.samples_drawn >= self.num_samples, "All samples should have been drawn before processing remaining candidates"
             with ThreadPoolExecutor(max_workers=self.join_parallelism) as executor:
                futures = []
                for left_candidate, right_candidate, embedding_sim in join_candidates:
                    # if the embedding similarity is lower than the threshold below which no records joined,
                    # then we can skip the LLM call and mark the records as not joined
                    if embedding_sim < self.max_non_matching_sim:
                        futures.append(executor.submit(self._process_join_candidate_with_sim, left_candidate, right_candidate, embedding_sim, passed_operator=False))

                    # if the embedding similarity is higher than the threshold above which all records joined,
                    # then we can skip the LLM call and mark the records as joined
                    elif embedding_sim > self.min_matching_sim:
                        futures.append(executor.submit(self._process_join_candidate_with_sim, left_candidate, right_candidate, embedding_sim, passed_operator=True))

                    # otherwise, we will process the LLM call
                    else:
                        futures.append(executor.submit(self._process_join_candidate_pair, left_candidate, right_candidate, gen_kwargs, embedding_sim))

                    num_inputs_processed += 1

                # collect results as they complete
                similarities, joined = [], []
                for future in as_completed(futures):
                    self.join_idx += 1
                    join_output_record, join_output_record_op_stats, embedding_sim = future.result()
                    output_records.append(join_output_record)
                    output_record_op_stats.append(join_output_record_op_stats)
                    similarities.append(embedding_sim)
                    joined.append(join_output_record._passed_operator)
                    print(f"{self.join_idx} JOINED")

                ### update thresholds if there are llm calls which incrementally squeeze the boundaries ###
                # sort join results by embedding similarity
                sorted_sim_join_tuples = sorted(zip(similarities, joined), key=lambda x: x[0])

                # potentially update threshold below which no records joined
                for embedding_sim, records_joined in sorted_sim_join_tuples:
                    if records_joined:
                        break
                    if not records_joined and embedding_sim > self.max_non_matching_sim:
                        self.max_non_matching_sim = embedding_sim

                # potentially update threshold above which all records joined
                for embedding_sim, records_joined in reversed(sorted_sim_join_tuples):
                    if not records_joined:
                        break
                    if records_joined and embedding_sim < self.min_matching_sim:
                        self.min_matching_sim = embedding_sim

        # amortize embedding costs over all output records and add to each record's op stats
        amortized_embedding_cost = total_embedding_cost / len(output_record_op_stats) if len(output_record_op_stats) > 0 else 0.0
        for record_op_stats in output_record_op_stats:
            record_op_stats.cost_per_record += amortized_embedding_cost
            record_op_stats.total_embedding_cost = amortized_embedding_cost

        # store input records to join with new records added later
        if self.retain_inputs:
            self._left_input_records.extend(zip(left_candidates, left_embeddings))
            self._right_input_records.extend(zip(right_candidates, right_embeddings))

        # if this is the final call, then add in any left/right/outer join records that did not match
        if final:
            return self._compute_unmatched_records(), 0

        # return empty DataRecordSet if no output records were produced
        if len(output_records) == 0:
            self.residual_embedding_cost = total_embedding_cost
            return DataRecordSet([], []), num_inputs_processed

        return DataRecordSet(output_records, output_record_op_stats), num_inputs_processed



import logging
from concurrent.futures import ThreadPoolExecutor, wait

from palimpzest.constants import PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.models import PlanStats
from palimpzest.query.execution.execution_strategy import ExecutionStrategy
from palimpzest.query.operators.aggregate import AggregateOp
from palimpzest.query.operators.distinct import DistinctOp
from palimpzest.query.operators.join import JoinOp
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.operators.scan import ContextScanOp, ScanPhysicalOp
from palimpzest.query.optimizer.plan import PhysicalPlan
from palimpzest.utils.progress import create_progress_manager

logger = logging.getLogger(__name__)


class ParallelExecutionStrategy(ExecutionStrategy):
    """
    A parallel execution strategy that processes data through a pipeline of operators using thread-based parallelism.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _any_queue_not_empty(self, queues: dict[str, list] | dict[str, dict[str, list]]) -> bool:
        """Helper function to check if any queue is not empty."""
        for _, value in queues.items():
            if isinstance(value, dict):
                if any(len(subqueue) > 0 for subqueue in value.values()):
                    return True
            elif len(value) > 0:
                return True
        return False

    def _upstream_ops_finished(self, plan: PhysicalPlan, unique_full_op_id: str, input_queues: dict[str, dict[str, list]], future_queues: dict[str, list]) -> bool:
        """Helper function to check if agg / join operator is ready to process its inputs."""
        upstream_unique_full_op_ids = plan.get_upstream_unique_full_op_ids(unique_full_op_id)
        upstream_input_queues = {upstream_unique_full_op_id: input_queues[upstream_unique_full_op_id] for upstream_unique_full_op_id in upstream_unique_full_op_ids}
        upstream_future_queues = {upstream_unique_full_op_id: future_queues[upstream_unique_full_op_id] for upstream_unique_full_op_id in upstream_unique_full_op_ids}
        return not (self._any_queue_not_empty(upstream_input_queues) or self._any_queue_not_empty(upstream_future_queues))

    def _finish_outer_join(self, executor: ThreadPoolExecutor, plan: PhysicalPlan, unique_full_op_id: str, input_queues: dict[str, dict[str, list]], future_queues: dict[str, list]) -> None:
        join_op_upstream_finished = self._upstream_ops_finished(plan, unique_full_op_id, input_queues, future_queues)
        join_input_queues_empty = all(len(inputs) == 0 for inputs in input_queues[unique_full_op_id].values())
        join_future_queue_empty = len(future_queues[unique_full_op_id]) == 0
        if join_op_upstream_finished and join_input_queues_empty and join_future_queue_empty:
            # process the join one last time with final=True to handle any left/right/outer join logic
            operator = self.unique_full_op_id_to_operator[unique_full_op_id]
            if not operator.finished:
                def finalize_op(operator):
                    return operator([], [], final=True)
                future = executor.submit(finalize_op, operator)
                future_queues[unique_full_op_id].append(future)
                operator.set_finished()

    def _process_future_results(self, unique_full_op_id: str, future_queues: dict[str, list], plan_stats: PlanStats) -> list[DataRecord]:
        """
        Helper function which takes a full operator id, the future queues, and plan stats, and performs
        the updates to plan stats and progress manager before returning the results from the finished futures.
        """
        # this function is called when the future queue is not empty
        # and the executor is not busy processing other futures
        done_futures, not_done_futures = wait(future_queues[unique_full_op_id], timeout=PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS)

        # add the unfinished futures back to the previous op's future queue
        future_queues[unique_full_op_id] = list(not_done_futures)

        # add the finished futures to the input queue for this operator
        output_records, total_inputs_processed, total_cost = [], 0, 0.0
        for future in done_futures:
            output = future.result()
            record_set, num_inputs_processed = output if self.is_join_op[unique_full_op_id] else (output, 1)

            # record set can be empty if one side of join has no input records yet
            if len(record_set) == 0:
                continue

            # otherwise, process records and their stats
            records = record_set.data_records
            record_op_stats = record_set.record_op_stats

            # update the inputs processed and total cost
            total_inputs_processed += num_inputs_processed
            total_cost += record_set.get_total_cost()

            # update plan stats
            plan_stats.add_record_op_stats(unique_full_op_id, record_op_stats)

            # add records which aren't filtered to the output records
            output_records.extend([record for record in records if record._passed_operator])

        # update the progress manager
        if total_inputs_processed > 0:
            num_outputs = len(output_records)
            self.progress_manager.incr(unique_full_op_id, num_inputs=total_inputs_processed, num_outputs=num_outputs, total_cost=total_cost)

        return output_records

    def _execute_plan(
            self,
            plan: PhysicalPlan,
            input_queues: dict[str, dict[str, list]],
            future_queues: dict[str, list],
            plan_stats: PlanStats,
        ) -> tuple[list[DataRecord], PlanStats]:
        # process all of the input records using a thread pool
        output_records = []
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            logger.debug(f"Created thread pool with {self.max_workers} workers")

            # execute the plan until either:
            # 1. all records have been processed, or
            # 2. the final limit operation has completed (we break out of the loop if this happens)
            final_op = plan.operator
            while self._any_queue_not_empty(input_queues) or self._any_queue_not_empty(future_queues):
                for topo_idx, operator in enumerate(plan):
                    source_unique_full_op_ids = (
                        [f"source_{operator.get_full_op_id()}"]
                        if isinstance(operator, (ContextScanOp, ScanPhysicalOp))
                        else plan.get_source_unique_full_op_ids(topo_idx, operator)
                    )
                    unique_full_op_id = f"{topo_idx}-{operator.get_full_op_id()}"

                    # get any finished futures from the previous operator and add them to the input queue for this operator
                    if not isinstance(operator, (ContextScanOp, ScanPhysicalOp)):
                        for source_unique_full_op_id in source_unique_full_op_ids:
                            records = self._process_future_results(source_unique_full_op_id, future_queues, plan_stats)
                            input_queues[unique_full_op_id][source_unique_full_op_id].extend(records)

                            # if the source is a left/right/outer join operator with no more inputs to process, then finish it
                            if self.is_outer_join_op[source_unique_full_op_id]:
                                self._finish_outer_join(executor, plan, source_unique_full_op_id, input_queues, future_queues)

                    # for the final operator, add any finished futures to the output_records
                    if unique_full_op_id == f"{topo_idx}-{final_op.get_full_op_id()}":
                        records = self._process_future_results(unique_full_op_id, future_queues, plan_stats)
                        output_records.extend(records)

                        # if this is a left/right/outer join operator with no more inputs to process, then finish it
                        if self.is_outer_join_op[unique_full_op_id]:
                            self._finish_outer_join(executor, plan, unique_full_op_id, input_queues, future_queues)

                    # if this operator does not have enough inputs to execute, then skip it
                    num_inputs = sum(len(inputs) for inputs in input_queues[unique_full_op_id].values())
                    agg_op_not_ready = isinstance(operator, AggregateOp) and not self._upstream_ops_finished(plan, unique_full_op_id, input_queues, future_queues)
                    join_op_not_ready = isinstance(operator, JoinOp) and not self.join_has_downstream_limit_op[unique_full_op_id] and not self._upstream_ops_finished(plan, unique_full_op_id, input_queues, future_queues)
                    if num_inputs == 0 or agg_op_not_ready or join_op_not_ready:
                        continue

                    # if this operator is an aggregate, process all the records in the input queue
                    if isinstance(operator, AggregateOp):
                        source_unique_full_op_id = source_unique_full_op_ids[0]
                        input_records = [input_queues[unique_full_op_id][source_unique_full_op_id].pop(0) for _ in range(num_inputs)]
                        future = executor.submit(operator, input_records)
                        future_queues[unique_full_op_id].append(future)

                    # if this operator is a join, process all pairs of records from the two input queues
                    elif isinstance(operator, JoinOp):
                        left_unique_full_source_op_id = source_unique_full_op_ids[0]
                        left_num_inputs = len(input_queues[unique_full_op_id][left_unique_full_source_op_id])
                        left_input_records = [input_queues[unique_full_op_id][left_unique_full_source_op_id].pop(0) for _ in range(left_num_inputs)]

                        right_unique_full_source_op_id = source_unique_full_op_ids[1]
                        right_num_inputs = len(input_queues[unique_full_op_id][right_unique_full_source_op_id])
                        right_input_records = [input_queues[unique_full_op_id][right_unique_full_source_op_id].pop(0) for _ in range(right_num_inputs)]

                        # NOTE: it would be nice to use executor for join inputs here; but for now synchronizing may be necessary
                        # future = executor.submit(operator, left_input_records, right_input_records)
                        # future_queues[unique_full_op_id].append(future)
                        record_set, num_inputs_processed = operator(left_input_records, right_input_records)
                        def no_op(rset, num_inputs_processed):
                            return rset, num_inputs_processed
                        future = executor.submit(no_op, record_set, num_inputs_processed)
                        future_queues[unique_full_op_id].append(future)

                    # if this operator is a limit, process one record at a time
                    elif isinstance(operator, LimitScanOp):
                        source_unique_full_op_id = source_unique_full_op_ids[0]
                        num_records_to_process = min(len(input_queues[unique_full_op_id][source_unique_full_op_id]), operator.limit - len(output_records))
                        for _ in range(num_records_to_process):
                            input_record = input_queues[unique_full_op_id][source_unique_full_op_id].pop(0)
                            future = executor.submit(operator, input_record)
                            future_queues[unique_full_op_id].append(future)

                        # if this is the final operator, add any finished futures to the output_records
                        # immediately so that we can break out of the loop if we've reached the limit
                        if unique_full_op_id == f"{topo_idx}-{final_op.get_full_op_id()}":
                            records = self._process_future_results(unique_full_op_id, future_queues, plan_stats)
                            output_records.extend(records)

                    # if this operator is a distinct, process records sequentially
                    # (distinct is not parallelized because it requires maintaining a set of seen records)
                    elif isinstance(operator, DistinctOp):
                        source_unique_full_op_id = source_unique_full_op_ids[0]
                        input_records = input_queues[unique_full_op_id][source_unique_full_op_id]
                        for record in input_records:
                            record_set = operator(record)
                            def no_op(rset):
                                return rset
                            future = executor.submit(no_op, record_set)
                            future_queues[unique_full_op_id].append(future)

                        # clear the input queue for this operator since we processed all records
                        input_queues[unique_full_op_id][source_unique_full_op_id].clear()

                    # otherwise, process records according to batch size
                    else:
                        source_unique_full_op_id = source_unique_full_op_ids[0]
                        input_records = input_queues[unique_full_op_id][source_unique_full_op_id]
                        if self.batch_size is None:
                            for input_record in input_records:
                                future = executor.submit(operator, input_record)
                                future_queues[unique_full_op_id].append(future)
                            input_queues[unique_full_op_id][source_unique_full_op_id].clear()
                        else:
                            batch_size = min(self.batch_size, len(input_records))
                            batch_input_records = input_records[:batch_size]
                            for input_record in batch_input_records:
                                future = executor.submit(operator, input_record)
                                future_queues[unique_full_op_id].append(future)
                            input_queues[unique_full_op_id][source_unique_full_op_id] = input_records[batch_size:]

                # TODO: change logic to stop upstream operators once a limit is reached
                # break out of loop if the final operator is a LimitScanOp and we've reached its limit
                if isinstance(final_op, LimitScanOp) and len(output_records) == final_op.limit:
                    break

        # finalize plan stats
        plan_stats.finish()

        return output_records, plan_stats

    def execute_plan(self, plan: PhysicalPlan):
        """Initialize the stats and execute the plan."""
        logger.info(f"Executing plan {plan.plan_id} with {self.max_workers} workers")
        logger.info(f"Plan Details: {plan}")

        # initialize plan stats
        plan_stats = PlanStats.from_plan(plan)
        plan_stats.start()

        # initialize input queues and future queues for each operation
        input_queues = self._create_input_queues(plan)
        future_queues = {f"{topo_idx}-{op.get_full_op_id()}": [] for topo_idx, op in enumerate(plan)}

        # precompute which operators are (outer) joins and which joins have downstream limit ops
        self.is_join_op = {f"{topo_idx}-{op.get_full_op_id()}": isinstance(op, JoinOp) for topo_idx, op in enumerate(plan)}
        self.is_outer_join_op = {f"{topo_idx}-{op.get_full_op_id()}": isinstance(op, JoinOp) and op.how in ("left", "right", "outer") for topo_idx, op in enumerate(plan)}
        self.join_has_downstream_limit_op = {}
        for topo_idx, op in enumerate(plan):
            if isinstance(op, JoinOp):
                unique_full_op_id = f"{topo_idx}-{op.get_full_op_id()}"
                has_downstream_limit_op = False
                for inner_topo_idx, op in enumerate(plan):
                    if inner_topo_idx <= topo_idx:
                        continue
                    if isinstance(op, LimitScanOp):
                        has_downstream_limit_op = True
                        break
                self.join_has_downstream_limit_op[unique_full_op_id] = has_downstream_limit_op

        # precompute mapping from unique_full_op_id to operator instance
        self.unique_full_op_id_to_operator = {f"{topo_idx}-{op.get_full_op_id()}": op for topo_idx, op in enumerate(plan)}

        # initialize and start the progress manager
        self.progress_manager = create_progress_manager(plan, num_samples=self.num_samples, progress=self.progress)
        self.progress_manager.start()

        # NOTE: we must handle progress manager outside of _execute_plan to ensure that it is shut down correctly;
        #       if we don't have the `finally:` branch, then program crashes can cause future program runs to fail
        #       because the progress manager cannot get a handle to the console 
        try:
            # execute plan
            output_records, plan_stats = self._execute_plan(plan, input_queues, future_queues, plan_stats)

        finally:
            # finish progress tracking
            self.progress_manager.finish()

        logger.info(f"Done executing plan: {plan.plan_id}")
        logger.debug(f"Plan stats: (plan_cost={plan_stats.total_plan_cost}, plan_time={plan_stats.total_plan_time})")

        return output_records, plan_stats




import logging

import numpy as np
from chromadb.api.models.Collection import Collection

from palimpzest.core.data.dataset import Dataset
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.models import OperatorCostEstimates, OperatorStats, RecordOpStats, SentinelPlanStats
from palimpzest.policy import Policy
from palimpzest.query.execution.execution_strategy import SentinelExecutionStrategy
from palimpzest.query.operators.aggregate import AggregateOp
from palimpzest.query.operators.convert import LLMConvert
from palimpzest.query.operators.filter import FilterOp, LLMFilter, NonLLMFilter
from palimpzest.query.operators.join import JoinOp
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.scan import ContextScanOp, ScanPhysicalOp
from palimpzest.query.operators.topk import TopKOp
from palimpzest.query.optimizer.plan import SentinelPlan
from palimpzest.utils.progress import create_progress_manager
from palimpzest.validator.validator import Validator

logger = logging.getLogger(__name__)

# NOTE: we currently do not support Sentinel Plans with aggregates or limits which are not the final plan operator

class OpFrontier:
    """
    This class represents the set of operators which are currently in the frontier for a given logical operator.
    Each operator in the frontier is an instance of a PhysicalOperator which either:

    1. lies on the Pareto frontier of the set of sampled operators, or
    2. has been sampled fewer than j times
    """

    def __init__(
            self,
            op_set: list[PhysicalOperator],
            source_unique_logical_op_ids: list[str],
            root_dataset_ids: list[str],
            source_indices: list[tuple],
            k: int,
            j: int,
            seed: int,
            policy: Policy,
            priors: dict | None = None,
            dont_use_priors: bool = False,
        ):
        # set k and j, which are the initial number of operators in the frontier and the
        # initial number of records to sample for each frontier operator
        self.k = min(k, len(op_set))
        self.j = j
        self.source_indices = source_indices
        self.root_dataset_ids = root_dataset_ids
        self.dont_use_priors = dont_use_priors

        # store the policy that we are optimizing under
        self.policy = policy

        # store the prior beliefs on operator performance (if provided)
        self.priors = priors

        # boolean indication of the type of operator in this OpFrontier
        sample_op = op_set[0]
        self.is_scan_op = isinstance(sample_op, (ScanPhysicalOp, ContextScanOp))
        self.is_filter_op = isinstance(sample_op, FilterOp)
        self.is_aggregate_op = isinstance(sample_op, AggregateOp)
        self.is_llm_join = isinstance(sample_op, JoinOp)
        is_llm_convert = isinstance(sample_op, LLMConvert)
        is_llm_filter = isinstance(sample_op, LLMFilter)
        is_llm_topk = isinstance(sample_op, TopKOp) and isinstance(sample_op.index, Collection)
        self.is_llm_op = is_llm_convert or is_llm_filter or is_llm_topk or self.is_llm_join
        self.is_llm_convert = is_llm_convert

        # get order in which we will sample physical operators for this logical operator
        sample_op_indices = self._get_op_index_order(op_set, seed)

        # construct the initial set of frontier and reservoir operators
        self.frontier_ops = [op_set[sample_idx] for sample_idx in sample_op_indices[:self.k]]
        self.reservoir_ops = [op_set[sample_idx] for sample_idx in sample_op_indices[self.k:]]
        self.off_frontier_ops: list[PhysicalOperator] = []

        # keep track of the source indices processed by each physical operator
        self.full_op_id_to_sources_processed = {op.get_full_op_id(): set() for op in op_set}
        self.full_op_id_to_sources_not_processed = {op.get_full_op_id(): source_indices for op in op_set}
        self.max_inputs = len(source_indices)

        # set the initial inputs for this logical operator; we maintain a mapping from source_unique_logical_op_id --> source_indices --> input;
        # for each unique source and (tuple of) source indices, we store its output, which is an input to this operator
        # for scan operators, we use the default name "source" since these operators have no source
        self.source_indices_to_inputs = {source_unique_logical_op_id: {} for source_unique_logical_op_id in source_unique_logical_op_ids}
        if self.is_scan_op:
            self.source_indices_to_inputs["source"] = {source_idx: [int(source_idx.split("-")[-1])] for source_idx in source_indices}
        

    def get_frontier_ops(self) -> list[PhysicalOperator]:
        """
        Returns the set of frontier operators for this OpFrontier.
        """
        return self.frontier_ops

    def get_off_frontier_ops(self) -> list[PhysicalOperator]:
        """
        Returns the set of off-frontier operators for this OpFrontier.
        """
        return self.off_frontier_ops

    def _compute_op_id_to_pareto_distance(self, priors: dict[str, dict[str, float]]) -> dict[str, float]:
        """
        Return l2-distance for each operator from the pareto frontier.
        """
        # get the dictionary representation of this poicy
        policy_dict = self.policy.get_dict()

        # compute the pareto optimal set of operators
        pareto_op_set = set()
        for op_id, metrics in priors.items():
            cost, time, quality = metrics["cost"], metrics["time"], metrics["quality"]
            pareto_frontier = True

            # check if any other operator dominates op_id
            for other_op_id, other_metrics in priors.items():
                other_cost, other_time, other_quality = other_metrics["cost"], other_metrics["time"], other_metrics["quality"]
                if op_id == other_op_id:
                    continue

                # if op_id is dominated by other_op_id, set pareto_frontier = False and break
                # NOTE: here we use a strict inequality (instead of the usual <= or >=) because
                #       all ops which have equal cost / time / quality / sel. should not be
                #       filtered out from sampling by our logic in this function
                cost_dominated = True if policy_dict["cost"] == 0.0 else other_cost < cost
                time_dominated = True if policy_dict["time"] == 0.0 else other_time < time
                quality_dominated = True if policy_dict["quality"] == 0.0 else other_quality > quality
                if cost_dominated and time_dominated and quality_dominated:
                    pareto_frontier = False
                    break

            # add op_id to pareto frontier if it's not dominated
            if pareto_frontier:
                pareto_op_set.add(op_id)

        # compute the shortest distance from each operator to the pareto frontier
        op_id_to_pareto_distance = {}
        for op_id, metrics in priors.items():
            # set distance to 0.0 if this operator is on the pareto frontier
            if op_id in pareto_op_set:
                op_id_to_pareto_distance[op_id] = 0.0
                continue

            # otherwise, compute min_dist to pareto operators
            min_dist = None
            cost, time, quality = metrics["cost"], metrics["time"], metrics["quality"]
            for pareto_op_id in pareto_op_set:
                pareto_cost, pareto_time, pareto_quality = priors[pareto_op_id]["cost"], priors[pareto_op_id]["time"], priors[pareto_op_id]["quality"]

                cost_dist_squared = 0.0 if policy_dict["cost"] == 0.0 else (cost - pareto_cost) ** 2
                time_dist_squared = 0.0 if policy_dict["time"] == 0.0 else (time - pareto_time) ** 2
                quality_dist_squared = 0.0 if policy_dict["quality"] == 0.0 else (quality - pareto_quality) ** 2
                dist = np.sqrt(cost_dist_squared + time_dist_squared + quality_dist_squared)
                if min_dist is None or dist < min_dist:
                    min_dist = dist

            # set minimum distance for this operator
            op_id_to_pareto_distance[op_id] = min_dist
        
        return op_id_to_pareto_distance

    def _compute_naive_priors(self, op_set: list[PhysicalOperator]) -> dict[str, dict[str, float]]:
        naive_priors = {}
        for op in op_set:
            # use naive cost estimates with dummy source estimates to compute priors
            source_op_estimates = OperatorCostEstimates(quality=1.0, cost_per_record=0.0, time_per_record=0.0, cardinality=100)
            op_estimates = (
                op.naive_cost_estimates(source_op_estimates, source_op_estimates)
                if self.is_llm_join
                else op.naive_cost_estimates(source_op_estimates)
            )

            # get op_id for this operator
            op_id = op.get_op_id()

            # set the naive quality, cost, and time priors for this operator
            naive_priors[op_id] = {
                "quality": op_estimates.quality,
                "cost": op_estimates.cost_per_record,
                "time": op_estimates.time_per_record,
            }

        return naive_priors

    def _get_op_index_order(self, op_set: list[PhysicalOperator], seed: int) -> list[int]:
        """
        Returns a list of indices for the operators in the op_set.
        """
        # if this is not an llm-operator, we simply return the indices in random order
        if not self.is_llm_op or self.dont_use_priors:
            if self.is_llm_convert:
                print("Using NO PRIORS for operator sampling order")
            rng = np.random.default_rng(seed=seed)
            op_indices = np.arange(len(op_set))
            rng.shuffle(op_indices)
            return op_indices

        # if this is an llm-operator, but we do not have priors, we first compute naive priors
        if self.priors is None or any([op_id not in self.priors for op_id in map(lambda op: op.get_op_id(), op_set)]):
            if self.is_llm_convert:
                print("Using NAIVE PRIORS for operator sampling order")
            self.priors = self._compute_naive_priors(op_set)

        # NOTE: self.priors is a dictionary with format:
        # {op_id: {"quality": quality, "cost": cost, "time": time}}

        # compute mean and std. dev. for each field
        qualities = [op_priors["quality"] for op_priors in self.priors.values()]
        costs = [op_priors["cost"] for op_priors in self.priors.values()]
        times = [op_priors["time"] for op_priors in self.priors.values()]
        metric_to_mean = {"quality": np.mean(qualities), "cost": np.mean(costs), "time": np.mean(times)}
        metric_to_std = {"quality": np.std(qualities), "cost": np.std(costs), "time": np.std(times)}

        # normalize the scale of each field to be the same
        for _, op_priors in self.priors.items():
            for metric, value in op_priors.items():
                if metric_to_std[metric] == 0.0:
                    op_priors[metric] = metric_to_mean[metric]
                else:
                    op_priors[metric] = (value - metric_to_mean[metric]) / metric_to_std[metric]

        # then, we compute the l2-distance from the pareto frontier for each operator
        op_id_to_distance = self._compute_op_id_to_pareto_distance(self.priors)

        # compute tuple for every operator, invert quality so ascending sort puts
        # best operator first: (op_id, dist, -1 * quality, cost, time);
        op_tuples = []
        for op in op_set:
            op_id = op.get_op_id()
            op_priors = self.priors[op_id]
            op_tuple = (op_id, op_id_to_distance[op_id], -1 * op_priors["quality"], op_priors["cost"], op_priors["time"])
            op_tuples.append(op_tuple)

        # sort tuples on distance, then second dim
        second_dim_idx = None
        if self.policy.get_primary_metric() == "quality":
            second_dim_idx = 2
        elif self.policy.get_primary_metric() == "cost":
            second_dim_idx = 3
        elif self.policy.get_primary_metric() == "time":
            second_dim_idx = 4

        # sort based on distance from pareto frontier; break ties with performance on max / min metric
        op_tuples = sorted(op_tuples, key=lambda x: (x[1], x[second_dim_idx]))

        # return final list of op indices in sample order
        op_id_to_idx = {op.get_op_id(): idx for idx, op in enumerate(op_set)}
        op_indices = [op_id_to_idx[op_tuple[0]] for op_tuple in op_tuples]

        return op_indices

    def _get_op_source_indices_pairs(self) -> list[tuple[PhysicalOperator, tuple[str] | None]]:
        """
        Returns a list of tuples for (op, source_indices) which this operator needs to execute
        in the next iteration.
        """
        op_source_indices_pairs = []

        # if this operator is not being optimized: we don't request inputs, but simply process what we are given / told to (in the case of scans)
        if not self.is_llm_op and len(self.frontier_ops) == 1:
            return [(self.frontier_ops[0], None)]

        # otherwise, sample (operator, source_indices) pairs
        for op in self.frontier_ops:
            # execute new operators on first j indices per root dataset, and previously sampled operators on one per root dataset
            new_operator = self.full_op_id_to_sources_processed[op.get_full_op_id()] == set()
            samples_per_root_dataset = self.j if new_operator else 1
            num_root_datasets = len(self.root_dataset_ids)
            num_samples = samples_per_root_dataset**num_root_datasets
            samples = self.full_op_id_to_sources_not_processed[op.get_full_op_id()][:num_samples]
            for source_indices in samples:
                op_source_indices_pairs.append((op, source_indices))

        return op_source_indices_pairs

    def get_source_indices_for_next_iteration(self) -> set[tuple[str]]:
        """
        Returns the set of source indices which need to be sampled for the next iteration.
        """
        op_source_indices_pairs = self._get_op_source_indices_pairs()
        return set([source_indices for _, source_indices in op_source_indices_pairs if source_indices is not None])

    def get_frontier_op_inputs(self, source_indices_to_sample: set[tuple[str]], max_quality_op: PhysicalOperator) -> list[tuple[PhysicalOperator, tuple[str], list[DataRecord] | list[int] | None]]:
        """
        Returns the list of frontier operators and their next input to process. If there are
        any indices in `source_indices_to_sample` which this operator does not sample on its own, then
        we also have this frontier process those source indices' input with its max quality operator.
        """
        # if this is an aggregate, run on every input
        if self.is_aggregate_op:
            # NOTE: we don't keep track of source indices for aggregate (would require computing powerset of all source records);
            #       thus, we cannot currently support optimizing plans w/LLM operators after aggregations
            op = self.frontier_ops[0]
            all_inputs = []
            for _, source_indices_to_inputs in self.source_indices_to_inputs.items():
                for _, inputs in source_indices_to_inputs.items():
                    all_inputs.extend(inputs)
            return [(op, tuple(), all_inputs)]

        ### for optimized operators
        # get the list of (op, source_indices) pairs which this operator needs to execute
        op_source_indices_pairs = self._get_op_source_indices_pairs()

        # remove any root datasets which this op frontier does not have access to from the source_indices_to_sample
        def remove_unavailable_root_datasets(source_indices: str | tuple) -> str | tuple | None:
            # base case: source_indices is a string
            if isinstance(source_indices, str):
                return source_indices if source_indices.split("---")[0] in self.root_dataset_ids else None

            # recursive case: source_indices is a tuple
            left_indices = source_indices[0]
            right_indices = source_indices[1]
            left_filtered = remove_unavailable_root_datasets(left_indices)
            right_filtered = remove_unavailable_root_datasets(right_indices)
            if left_filtered is None and right_filtered is None:
                return None

            if left_filtered is None:
                return right_filtered
            if right_filtered is None:
                return left_filtered
            return (left_filtered, right_filtered)

        source_indices_to_sample = {remove_unavailable_root_datasets(source_indices) for source_indices in source_indices_to_sample}

        # if there are any source_indices in source_indices_to_sample which are not sampled by this operator,
        # apply the max quality operator (and any other frontier operators with no samples)
        sampled_source_indices = set(map(lambda tup: tup[1], op_source_indices_pairs))
        unsampled_source_indices = source_indices_to_sample - sampled_source_indices
        for source_indices in unsampled_source_indices:
            op_source_indices_pairs.append((max_quality_op, source_indices))
            for op in self.frontier_ops:
                if self.full_op_id_to_sources_processed[op.get_full_op_id()] == set() and op.get_full_op_id() != max_quality_op.get_full_op_id():
                    op_source_indices_pairs.append((op, source_indices))

        # construct the op inputs
        op_inputs = []
        if self.is_llm_join:
            left_source_unique_logical_op_id, right_source_unique_logical_op_id = list(self.source_indices_to_inputs)
            left_source_indices_to_inputs = self.source_indices_to_inputs[left_source_unique_logical_op_id]
            right_source_indices_to_inputs = self.source_indices_to_inputs[right_source_unique_logical_op_id]
            for op, source_indices in op_source_indices_pairs:
                left_source_indices = source_indices[0]
                right_source_indices = source_indices[1]
                left_inputs = left_source_indices_to_inputs.get(left_source_indices, [])
                right_inputs = right_source_indices_to_inputs.get(right_source_indices, [])
                if len(left_inputs) > 0 and len(right_inputs) > 0:
                    op_inputs.append((op, (left_source_indices, right_source_indices), (left_inputs, right_inputs)))
            return op_inputs

        # if operator is not a join
        source_unique_logical_op_id = list(self.source_indices_to_inputs)[0]
        op_inputs = [
            (op, source_indices, input)
            for op, source_indices in op_source_indices_pairs
            for input in self.source_indices_to_inputs[source_unique_logical_op_id].get(source_indices, [])
        ]

        return op_inputs

    def update_frontier(self, unique_logical_op_id: str, plan_stats: SentinelPlanStats, full_op_id_to_source_indices_processed: dict[str, set[list]]) -> None:
        """
        Update the set of frontier operators, pulling in new ones from the reservoir as needed.
        This function will:
        1. Compute the mean, LCB, and UCB for the cost, time, quality, and selectivity of each frontier operator
        2. Compute the pareto optimal set of frontier operators (using the mean values)
        3. Update the frontier and reservoir sets of operators based on their LCB/UCB overlap with the pareto frontier
        """
        # NOTE: downstream operators may end up re-computing the same record_id with a diff. input as upstream
        #       upstream operators change; in this case, we de-duplicate record_op_stats with identical record_ids
        #       and keep the one with the maximum quality
        # get a mapping from full_op_id --> list[RecordOpStats]
        full_op_id_to_op_stats: dict[str, OperatorStats] = plan_stats.operator_stats.get(unique_logical_op_id, {})
        full_op_id_to_record_op_stats: dict[str, list[RecordOpStats]] = {}
        for full_op_id, op_stats in full_op_id_to_op_stats.items():
            # skip over operators which have not been sampled
            if len(op_stats.record_op_stats_lst) == 0:
                continue

            # compute mapping from record_id to highest quality record op stats
            record_id_to_max_quality_record_op_stats = {}
            for record_op_stats in op_stats.record_op_stats_lst:
                record_id = record_op_stats.record_id
                if record_id not in record_id_to_max_quality_record_op_stats:  # noqa: SIM114
                    record_id_to_max_quality_record_op_stats[record_id] = record_op_stats

                elif record_op_stats.quality > record_id_to_max_quality_record_op_stats[record_id].quality:
                    record_id_to_max_quality_record_op_stats[record_id] = record_op_stats

            # compute final list of record op stats
            full_op_id_to_record_op_stats[full_op_id] = list(record_id_to_max_quality_record_op_stats.values())

        # NOTE: it is possible for the full_op_id_to_record_op_stats to be empty if there is a duplicate operator
        # (e.g. a scan of the same dataset) which has all of its results cached and no new_record_op_stats;
        # in this case, we do not update the frontier
        if full_op_id_to_record_op_stats == {}:
            return

        # update the set of source indices processed by each physical operator
        for full_op_id, source_indices_processed in full_op_id_to_source_indices_processed.items():
            # update the set of source indices processed
            for source_indices in source_indices_processed:
                source_indices = source_indices[0] if len(source_indices) == 1 else tuple(source_indices)
                self.full_op_id_to_sources_processed[full_op_id].add(source_indices)
                if source_indices in self.full_op_id_to_sources_not_processed[full_op_id]:
                    self.full_op_id_to_sources_not_processed[full_op_id].remove(source_indices)

            # update the set of source indices not processed
            self.full_op_id_to_sources_not_processed[full_op_id] = [
                indices for indices in self.full_op_id_to_sources_not_processed[full_op_id]
                if indices not in source_indices_processed
            ]

        # compute mapping of physical op to num samples and total samples drawn
        full_op_id_to_num_samples, total_num_samples = {}, 0
        for full_op_id, record_op_stats_lst in full_op_id_to_record_op_stats.items():
            # compute the number of samples as the length of the record_op_stats_lst
            num_samples = len(record_op_stats_lst)
            full_op_id_to_num_samples[full_op_id] = num_samples
            total_num_samples += num_samples

        # compute avg. selectivity, cost, time, and quality for each physical operator
        def total_output(record_op_stats_lst: list[RecordOpStats]):
            return sum([record_op_stats.passed_operator for record_op_stats in record_op_stats_lst])

        def total_input(record_op_stats_lst: list[RecordOpStats]):
            # TODO: this is okay for now because we only really need these calculations for Converts and Filters,
            #       but this will need more thought if/when we optimize joins
            all_parent_ids = []
            for record_op_stats in record_op_stats_lst:
                all_parent_ids.extend(
                    [None]
                    if record_op_stats.record_parent_ids is None
                    else record_op_stats.record_parent_ids
                )
            return len(set(all_parent_ids))

        full_op_id_to_mean_selectivity = {
            full_op_id: total_output(record_op_stats_lst) / total_input(record_op_stats_lst)
            for full_op_id, record_op_stats_lst in full_op_id_to_record_op_stats.items()
        }
        full_op_id_to_mean_cost = {
            full_op_id: np.mean([record_op_stats.cost_per_record for record_op_stats in record_op_stats_lst])
            for full_op_id, record_op_stats_lst in full_op_id_to_record_op_stats.items()
        }
        full_op_id_to_mean_time = {
            full_op_id: np.mean([record_op_stats.time_per_record for record_op_stats in record_op_stats_lst])
            for full_op_id, record_op_stats_lst in full_op_id_to_record_op_stats.items()
        }
        full_op_id_to_mean_quality = {
            full_op_id: np.mean([record_op_stats.quality for record_op_stats in record_op_stats_lst if record_op_stats.quality is not None])
            for full_op_id, record_op_stats_lst in full_op_id_to_record_op_stats.items()
        }

        # # compute average, LCB, and UCB of each operator; the confidence bounds depend upon
        # # the computation of the alpha parameter, which we scale to be 0.5 * the mean (of means)
        # # of the metric across all operators in this operator set
        # cost_alpha = 0.5 * np.mean([mean_cost for mean_cost in full_op_id_to_mean_cost.values()])
        # time_alpha = 0.5 * np.mean([mean_time for mean_time in full_op_id_to_mean_time.values()])
        # quality_alpha = 0.5 * np.mean([mean_quality for mean_quality in full_op_id_to_mean_quality.values()])
        # selectivity_alpha = 0.5 * np.mean([mean_selectivity for mean_selectivity in full_op_id_to_mean_selectivity.values()])
        cost_alpha = 0.5 * (np.max(list(full_op_id_to_mean_cost.values())) - np.min(list(full_op_id_to_mean_cost.values())))
        time_alpha = 0.5 * (np.max(list(full_op_id_to_mean_time.values())) - np.min(list(full_op_id_to_mean_time.values())))
        quality_alpha = 0.5 * (np.max(list(full_op_id_to_mean_quality.values())) - np.min(list(full_op_id_to_mean_quality.values())))
        selectivity_alpha = 0.5 * (np.max(list(full_op_id_to_mean_selectivity.values())) - np.min(list(full_op_id_to_mean_selectivity.values())))

        # compute metrics for each physical operator
        op_metrics = {}
        for full_op_id in full_op_id_to_record_op_stats:
            sample_ratio = np.sqrt(np.log(total_num_samples) / full_op_id_to_num_samples[full_op_id])
            exploration_terms = np.array([cost_alpha * sample_ratio, time_alpha * sample_ratio, quality_alpha * sample_ratio, selectivity_alpha * sample_ratio])
            mean_terms = (full_op_id_to_mean_cost[full_op_id], full_op_id_to_mean_time[full_op_id], full_op_id_to_mean_quality[full_op_id], full_op_id_to_mean_selectivity[full_op_id])

            # NOTE: we could clip these; however I will not do so for now to allow for arbitrary quality metric(s)
            lcb_terms = mean_terms - exploration_terms
            ucb_terms = mean_terms + exploration_terms
            op_metrics[full_op_id] = {"mean": mean_terms, "lcb": lcb_terms, "ucb": ucb_terms}

        # get the tuple representation of this policy
        policy_dict = self.policy.get_dict()

        # compute the pareto optimal set of operators
        pareto_op_set = set()
        for full_op_id, metrics in op_metrics.items():
            cost, time, quality, selectivity = metrics["mean"]
            pareto_frontier = True

            # check if any other operator dominates full_op_id
            for other_full_op_id, other_metrics in op_metrics.items():
                other_cost, other_time, other_quality, other_selectivity = other_metrics["mean"]
                if full_op_id == other_full_op_id:
                    continue

                # if full_op_id is dominated by other_full_op_id, set pareto_frontier = False and break
                # NOTE: here we use a strict inequality (instead of the usual <= or >=) because
                #       all ops which have equal cost / time / quality / sel. should not be
                #       filtered out from sampling by our logic in this function
                cost_dominated = True if policy_dict["cost"] == 0.0 else other_cost < cost
                time_dominated = True if policy_dict["time"] == 0.0 else other_time < time
                quality_dominated = True if policy_dict["quality"] == 0.0 else other_quality > quality
                selectivity_dominated = True if not self.is_filter_op else other_selectivity < selectivity
                if cost_dominated and time_dominated and quality_dominated and selectivity_dominated:
                    pareto_frontier = False
                    break

            # add full_op_id to pareto frontier if it's not dominated
            if pareto_frontier:
                pareto_op_set.add(full_op_id)

        # iterate over op metrics and compute the new frontier set of operators
        new_frontier_full_op_ids = set()
        for full_op_id, metrics in op_metrics.items():

            # if this op is fully sampled, do not keep it on the frontier
            if len(self.full_op_id_to_sources_processed[full_op_id]) == self.max_inputs:
                continue

            # if this op is pareto optimal keep it in our frontier ops
            if full_op_id in pareto_op_set:
                new_frontier_full_op_ids.add(full_op_id)
                continue

            # otherwise, if this op overlaps with an op on the pareto frontier, keep it in our frontier ops
            # NOTE: for now, we perform an optimistic comparison with the ucb/lcb
            pareto_frontier = True
            op_cost, op_time, _, op_selectivity = metrics["lcb"]
            op_quality = metrics["ucb"][2]
            for pareto_full_op_id in pareto_op_set:
                pareto_cost, pareto_time, _, pareto_selectivity = op_metrics[pareto_full_op_id]["ucb"]
                pareto_quality = op_metrics[pareto_full_op_id]["lcb"][2]

                # if full_op_id is dominated by pareto_full_op_id, set pareto_frontier = False and break
                cost_dominated = True if policy_dict["cost"] == 0.0 else pareto_cost <= op_cost
                time_dominated = True if policy_dict["time"] == 0.0 else pareto_time <= op_time
                quality_dominated = True if policy_dict["quality"] == 0.0 else pareto_quality >= op_quality
                selectivity_dominated = True if not self.is_filter_op else pareto_selectivity <= op_selectivity
                if cost_dominated and time_dominated and quality_dominated and selectivity_dominated:
                    pareto_frontier = False
                    break

            # add full_op_id to pareto frontier if it's not dominated
            if pareto_frontier:
                new_frontier_full_op_ids.add(full_op_id)

        # for operators that were in the frontier, keep them in the frontier if they
        # are still pareto optimal, otherwise, move them to the end of the reservoir
        new_frontier_ops = []
        for op in self.frontier_ops:
            if op.get_full_op_id() in new_frontier_full_op_ids:
                new_frontier_ops.append(op)
            else:
                self.off_frontier_ops.append(op)

        # if there are operators we previously sampled which are now back on the frontier
        # add them to the frontier, otherwise, put them back in the off_frontier_ops
        new_off_frontier_ops = []
        for op in self.off_frontier_ops:
            if op.get_full_op_id() in new_frontier_full_op_ids:
                new_frontier_ops.append(op)
            else:
                new_off_frontier_ops.append(op)

        # finally, if we have fewer than k operators in the frontier, sample new operators
        # from the reservoir and put them in the frontier
        while len(new_frontier_ops) < self.k and len(self.reservoir_ops) > 0:
            new_op = self.reservoir_ops.pop(0)
            new_frontier_ops.append(new_op)

        # update the frontier and off frontier ops
        self.frontier_ops = new_frontier_ops
        self.off_frontier_ops = new_off_frontier_ops

    def pick_highest_quality_output(self, record_sets: list[DataRecordSet]) -> DataRecordSet:
        # if there's only one operator in the set, we return its record_set
        if len(record_sets) == 1:
            return record_sets[0]

        # NOTE: I don't like that this assumes the models are consistent in
        #       how they order their record outputs for one-to-many converts;
        #       eventually we can try out more robust schemes to account for
        #       differences in ordering
        # aggregate records at each index in the response
        idx_to_records = {}
        for record_set in record_sets:
            for idx in range(len(record_set)):
                record, record_op_stats = record_set[idx], record_set.record_op_stats[idx]
                if idx not in idx_to_records:
                    idx_to_records[idx] = [(record, record_op_stats)]
                else:
                    idx_to_records[idx].append((record, record_op_stats))

        # compute highest quality answer at each index
        out_records = []
        out_record_op_stats = []
        for idx in range(len(idx_to_records)):
            records_lst, record_op_stats_lst = zip(*idx_to_records[idx])
            max_quality_record, max_quality = records_lst[0], record_op_stats_lst[0].quality if record_op_stats_lst[0].quality is not None else 0.0
            max_quality_stats = record_op_stats_lst[0]
            for record, record_op_stats in zip(records_lst[1:], record_op_stats_lst[1:]):
                record_quality = record_op_stats.quality if record_op_stats.quality is not None else 0.0
                if record_quality > max_quality:
                    max_quality_record = record
                    max_quality = record_quality
                    max_quality_stats = record_op_stats
            out_records.append(max_quality_record)
            out_record_op_stats.append(max_quality_stats)

        # create and return final DataRecordSet
        return DataRecordSet(out_records, out_record_op_stats)

    def update_inputs(self, source_unique_logical_op_id: str, source_indices_to_record_sets: dict[tuple[int], list[DataRecordSet]]):
        """
        Update the inputs for this logical operator based on the outputs of the previous logical operator.
        """
        for source_indices, record_sets in source_indices_to_record_sets.items():
            input = []
            max_quality_record_set = self.pick_highest_quality_output(record_sets)
            for record in max_quality_record_set:
                input.append(record if record._passed_operator else None)

            self.source_indices_to_inputs[source_unique_logical_op_id][source_indices] = input

class MABExecutionStrategy(SentinelExecutionStrategy):
    """
    This class implements the Multi-Armed Bandit (MAB) execution strategy for SentinelQueryProcessors.

    NOTE: the number of samples will slightly exceed the sample_budget if the number of operator
    calls does not perfectly match the sample_budget. This may cause some minor discrepancies with
    the progress manager as a result.
    """
    def _remove_filtered_records_from_downstream_ops(self, topo_idx: int, plan: SentinelPlan, op_frontiers: dict[str, OpFrontier], source_indices_to_all_record_sets: dict[int, list[DataRecordSet]]) -> None:
        """Remove records which were filtered out by a NonLLMFilter from all downstream operators."""
        filtered_source_indices = set()

        # NonLLMFilter will have one record_set per source_indices with a single record
        for source_indices, record_sets in source_indices_to_all_record_sets.items():
            record: DataRecord = record_sets[0][0]
            if not record._passed_operator:
                filtered_source_indices.add(source_indices)

        # remove filtered source indices from all downstream operators
        if len(filtered_source_indices) > 0:
            for downstream_topo_idx in range(topo_idx + 1, len(plan)):
                downstream_logical_op_id = plan[downstream_topo_idx][0]
                downstream_unique_logical_op_id = f"{downstream_topo_idx}-{downstream_logical_op_id}"
                downstream_op_frontier = op_frontiers[downstream_unique_logical_op_id]
                for full_op_id in downstream_op_frontier.full_op_id_to_sources_not_processed:
                    downstream_op_frontier.full_op_id_to_sources_not_processed[full_op_id] = [
                        indices for indices in downstream_op_frontier.full_op_id_to_sources_not_processed[full_op_id]
                        if indices not in filtered_source_indices
                    ]

    def _get_max_quality_op(self, unique_logical_op_id: str, op_frontiers: dict[str, OpFrontier], plan_stats: SentinelPlanStats) -> PhysicalOperator:
        """
        Returns the operator in the frontier with the highest (estimated) quality.
        """
        # get the (off) frontier operators for this logical_op_id
        frontier_ops = op_frontiers[unique_logical_op_id].get_frontier_ops() + op_frontiers[unique_logical_op_id].get_off_frontier_ops()

        # get a mapping from full_op_id --> list[RecordOpStats]
        full_op_id_to_op_stats: dict[str, OperatorStats] = plan_stats.operator_stats.get(unique_logical_op_id, {})
        full_op_id_to_record_op_stats = {
            full_op_id: op_stats.record_op_stats_lst
            for full_op_id, op_stats in full_op_id_to_op_stats.items()
        }

        # iterate over the frontier ops and return the one with the highest quality
        max_quality_op, max_avg_quality = None, None
        for op in frontier_ops:
            op_quality_stats = []
            if op.get_full_op_id() in full_op_id_to_record_op_stats:
                op_quality_stats = [
                    record_op_stats.quality
                    for record_op_stats in full_op_id_to_record_op_stats[op.get_full_op_id()]
                    if record_op_stats.quality is not None
                ]
            avg_op_quality = sum(op_quality_stats) / len(op_quality_stats) if len(op_quality_stats) > 0 else 0.0
            if max_avg_quality is None or avg_op_quality > max_avg_quality:
                max_quality_op = op
                max_avg_quality = avg_op_quality

        return max_quality_op

    def _compute_termination_condition(self, samples_drawn: int, sampling_cost: float) -> bool:
        return (samples_drawn >= self.sample_budget) if self.sample_cost_budget is None else (sampling_cost >= self.sample_cost_budget)

    def _execute_sentinel_plan(
            self,
            plan: SentinelPlan,
            op_frontiers: dict[str, OpFrontier],
            validator: Validator,
            plan_stats: SentinelPlanStats,
        ) -> SentinelPlanStats:
        # sample records and operators and update the frontiers
        samples_drawn, sampling_cost = 0, 0.0
        while not self._compute_termination_condition(samples_drawn, sampling_cost):
            # pre-compute the set of source indices which will need to be sampled
            source_indices_to_sample = set()
            for op_frontier in op_frontiers.values():
                source_indices = op_frontier.get_source_indices_for_next_iteration()
                source_indices_to_sample.update(source_indices)

            # execute operator sets in sequence
            for topo_idx, (logical_op_id, op_set) in enumerate(plan):
                # compute unique logical op id within plan
                unique_logical_op_id = f"{topo_idx}-{logical_op_id}"

                # use the execution cache to determine the maximum quality operator for this logical_op_id
                max_quality_op = self._get_max_quality_op(unique_logical_op_id, op_frontiers, plan_stats)

                # get frontier ops and their next input
                def filter_and_clean_inputs(frontier_op_inputs: list[tuple]) -> bool:
                    cleaned_inputs = []
                    for tup in frontier_op_inputs:
                        input = tup[-1]
                        if isinstance(input, list):
                            input = [record for record in input if record is not None]
                        if input is not None and input != []:
                            cleaned_inputs.append((tup[0], tup[1], input))
                    return cleaned_inputs
                frontier_op_inputs = op_frontiers[unique_logical_op_id].get_frontier_op_inputs(source_indices_to_sample, max_quality_op)
                frontier_op_inputs = filter_and_clean_inputs(frontier_op_inputs)

                # break out of the loop if frontier_op_inputs is empty, as this means all records have been filtered out
                if len(frontier_op_inputs) == 0:
                    continue

                # run sampled operators on sampled inputs and update the number of samples drawn
                source_indices_to_record_set_tuples, num_llm_ops = self._execute_op_set(unique_logical_op_id, frontier_op_inputs)
                samples_drawn += num_llm_ops

                # score the quality of each generated output
                source_indices_to_all_record_sets = {
                    source_indices: [(record_set, op) for record_set, op, _ in record_set_tuples]
                    for source_indices, record_set_tuples in source_indices_to_record_set_tuples.items()
                }
                source_indices_to_all_record_sets, val_gen_stats = self._score_quality(validator, source_indices_to_all_record_sets)

                # update the progress manager with validation cost
                self.progress_manager.incr_overall_progress_cost(val_gen_stats.cost_per_record)

                # remove records that were read from the execution cache before adding to record op stats
                new_record_op_stats = []
                for _, record_set_tuples in source_indices_to_record_set_tuples.items():
                    for record_set, _, is_new in record_set_tuples:
                        if is_new:
                            new_record_op_stats.extend(record_set.record_op_stats)

                # update plan stats
                plan_stats.add_record_op_stats(unique_logical_op_id, new_record_op_stats)
                plan_stats.add_validation_gen_stats(unique_logical_op_id, val_gen_stats)
                sampling_cost = plan_stats.get_total_cost_so_far()

                # provide the best record sets as inputs to the next logical operator
                next_unique_logical_op_id = plan.get_next_unique_logical_op_id(unique_logical_op_id)
                if next_unique_logical_op_id is not None:
                    source_indices_to_all_record_sets = {
                        source_indices: [record_set for record_set, _ in record_set_tuples]
                        for source_indices, record_set_tuples in source_indices_to_all_record_sets.items()
                    }
                    op_frontiers[next_unique_logical_op_id].update_inputs(unique_logical_op_id, source_indices_to_all_record_sets)

                # update the (pareto) frontier for each set of operators
                full_op_id_to_source_indices_processed = {}
                for source_indices, record_set_tuples in source_indices_to_record_set_tuples.items():
                    for _, op, _ in record_set_tuples:
                        if op.get_full_op_id() not in full_op_id_to_source_indices_processed:
                            full_op_id_to_source_indices_processed[op.get_full_op_id()] = set()
                        full_op_id_to_source_indices_processed[op.get_full_op_id()].add(source_indices)
                op_frontiers[unique_logical_op_id].update_frontier(unique_logical_op_id, plan_stats, full_op_id_to_source_indices_processed)

                # if the operator is a non-llm filter which has filtered out records, remove those records from
                # all downstream operators' full_op_id_to_sources_not_processed
                if isinstance(op_set[0], NonLLMFilter) and next_unique_logical_op_id is not None:
                    self._remove_filtered_records_from_downstream_ops(topo_idx, plan, op_frontiers, source_indices_to_all_record_sets)

        # finalize plan stats
        plan_stats.finish()

        return plan_stats


    def execute_sentinel_plan(self, plan: SentinelPlan, train_dataset: dict[str, Dataset], validator: Validator) -> SentinelPlanStats:
        logger.info(f"Executing plan {plan.plan_id} with {self.max_workers} workers")

        # initialize plan stats
        plan_stats = SentinelPlanStats.from_plan(plan)
        plan_stats.start()

        # shuffle the indices of records to sample
        dataset_id_to_shuffled_source_indices = {}
        for dataset_id, dataset in train_dataset.items():
            total_num_samples = len(dataset)
            shuffled_source_indices = [f"{dataset_id}---{int(idx)}" for idx in np.arange(total_num_samples)]
            self.rng.shuffle(shuffled_source_indices)
            dataset_id_to_shuffled_source_indices[dataset_id] = shuffled_source_indices

        # initialize frontier for each logical operator
        op_frontiers = {}
        for topo_idx, (logical_op_id, op_set) in enumerate(plan):
            unique_logical_op_id = f"{topo_idx}-{logical_op_id}"
            source_unique_logical_op_ids = plan.get_source_unique_logical_op_ids(unique_logical_op_id)
            root_dataset_ids = plan.get_root_dataset_ids(unique_logical_op_id)
            sample_op = op_set[0]
            if isinstance(sample_op, (ScanPhysicalOp, ContextScanOp)):
                assert len(root_dataset_ids) == 1, f"Scan for {sample_op} has {len(root_dataset_ids)} > 1 root dataset ids"
                root_dataset_id = root_dataset_ids[0]
                source_indices = dataset_id_to_shuffled_source_indices[root_dataset_id]
                op_frontiers[unique_logical_op_id] = OpFrontier(op_set, source_unique_logical_op_ids, root_dataset_ids, source_indices, self.k, self.j, self.seed, self.policy, self.priors, self.dont_use_priors)
            elif isinstance(sample_op, JoinOp):
                assert len(source_unique_logical_op_ids) == 2, f"Join for {sample_op} has {len(source_unique_logical_op_ids)} != 2 source logical operators"
                left_source_indices = op_frontiers[source_unique_logical_op_ids[0]].source_indices
                right_source_indices = op_frontiers[source_unique_logical_op_ids[1]].source_indices
                source_indices = []
                for left_source_idx in left_source_indices:
                    for right_source_idx in right_source_indices:
                        source_indices.append((left_source_idx, right_source_idx))
                op_frontiers[unique_logical_op_id] = OpFrontier(op_set, source_unique_logical_op_ids, root_dataset_ids, source_indices, self.k, self.j, self.seed, self.policy, self.priors, self.dont_use_priors)
            else:
                source_indices = op_frontiers[source_unique_logical_op_ids[0]].source_indices
                op_frontiers[unique_logical_op_id] = OpFrontier(op_set, source_unique_logical_op_ids, root_dataset_ids, source_indices, self.k, self.j, self.seed, self.policy, self.priors, self.dont_use_priors)

        # initialize and start the progress manager
        self.progress_manager = create_progress_manager(plan, sample_budget=self.sample_budget, sample_cost_budget=self.sample_cost_budget, progress=self.progress)
        self.progress_manager.start()

        # NOTE: we must handle progress manager outside of _execute_sentinel_plan to ensure that it is shut down correctly;
        #       if we don't have the `finally:` branch, then program crashes can cause future program runs to fail because
        #       the progress manager cannot get a handle to the console 
        try:
            # execute sentinel plan by sampling records and operators
            plan_stats = self._execute_sentinel_plan(plan, op_frontiers, validator, plan_stats)

        finally:
            # finish progress tracking
            self.progress_manager.finish()

        logger.info(f"Done executing sentinel plan: {plan.plan_id}")
        logger.debug(f"Plan stats: (plan_cost={plan_stats.total_plan_cost}, plan_time={plan_stats.total_plan_time})")

        return plan_stats



import os

from palimpzest.constants import Model


def get_models(include_embedding: bool = False, use_vertex: bool = False, gemini_credentials_path: str | None = None, api_base: str | None = None) -> list[Model]:
    """
    Return the set of models which the system has access to based on the set environment variables.
    """
    models = []
    if os.getenv("OPENAI_API_KEY") not in [None, ""]:
        openai_models = [model for model in Model if model.is_openai_model()]
        if not include_embedding:
            openai_models = [
                model for model in openai_models if not model.is_embedding_model()
            ]
        models.extend(openai_models)

    if os.getenv("TOGETHER_API_KEY") not in [None, ""]:
        together_models = [model for model in Model if model.is_together_model()]
        if not include_embedding:
            together_models = [
                model for model in together_models if not model.is_embedding_model()
            ]
        models.extend(together_models)

    if os.getenv("ANTHROPIC_API_KEY") not in [None, ""]:
        anthropic_models = [model for model in Model if model.is_anthropic_model()]
        if not include_embedding:
            anthropic_models = [
                model for model in anthropic_models if not model.is_embedding_model()
            ]
        models.extend(anthropic_models)

    gemini_credentials_path = (
        os.path.join(os.path.expanduser("~"), ".config", "gcloud", "application_default_credentials.json")
        if gemini_credentials_path is None
        else gemini_credentials_path
    )
    if os.getenv("GEMINI_API_KEY") not in [None, ""] or (use_vertex and os.path.exists(gemini_credentials_path)):
        vertex_models = [model for model in Model if model.is_vertex_model()]
        google_ai_studio_models = [model for model in Model if model.is_google_ai_studio_model()]
        if not include_embedding:
            vertex_models = [
                model for model in vertex_models if not model.is_embedding_model()
            ]
        if use_vertex:
            models.extend(vertex_models)
        else:
            models.extend(google_ai_studio_models)

    if api_base is not None:
        vllm_models = [model for model in Model if model.is_vllm_model()]
        if not include_embedding:
            vllm_models = [
                model for model in vllm_models if not model.is_embedding_model()
            ]
        models.extend(vllm_models)

    return models


def resolve_reasoning_settings(model: Model | None, reasoning_effort: str | None) -> tuple[bool, str]:
    """
    Resolve the reasoning settings based on the model and provided reasoning effort.
    Returns a tuple indicating whether reasoning prompt should be used and the reasoning effort level.
    By default, we use the reasoning prompt everywhere while setting the model reasoning effort to None (or minimal).
    If a user explicitly provides a reasoning_effort, we pass that through to the model.
    If the user explicitly disables reasoning_effort, we disable the reasoning prompt as well.
    """
    # turn off reasoning prompt if reasoning_effort is in [None, "disable", "minimal", "low"]
    use_reasoning_prompt = reasoning_effort not in [None, "disable", "minimal", "low"]

    # if reasoning_effort is set to "default", set it to None to use model defaults
    if reasoning_effort == "default":
        reasoning_effort = None

    # translate reasoning_effort into model-specific settings
    if model is not None and model.is_reasoning_model():
        if model.is_vertex_model() or model.is_google_ai_studio_model():
            if reasoning_effort is None and model in [Model.GEMINI_2_5_PRO, Model.GOOGLE_GEMINI_2_5_PRO]:
                reasoning_effort = "low"
            elif reasoning_effort is None:
                reasoning_effort = "disable"
        elif model.is_openai_model():
            reasoning_effort = "minimal" if reasoning_effort in [None, "disable", "minimal", "low"] else reasoning_effort

    return use_reasoning_prompt, reasoning_effort



from __future__ import annotations

import json
from typing import Callable

from pydantic import BaseModel

from palimpzest.constants import AggFunc, Cardinality
from palimpzest.core.data import context, dataset
from palimpzest.core.elements.filters import Filter
from palimpzest.core.elements.groupbysig import GroupBySig
from palimpzest.core.lib.schemas import Average, Count, Max, Min, Sum
from palimpzest.utils.hash_helpers import hash_for_id


class LogicalOperator:
    """
    A logical operator is an operator that operates on Sets.

    Right now it can be one of:
    - BaseScan (scans data from a root Dataset)
    - ContextScan (loads the context for a root Dataset)
    - FilteredScan (scans input Set and applies filter)
    - ConvertScan (scans input Set and converts it to new Schema)
    - LimitScan (scans up to N records from a Set)
    - GroupByAggregate (applies a group by on the Set)
    - Aggregate (applies an aggregation on the Set)
    - TopKScan (fetches documents from a provided input for a given query)
    - Map (applies a function to each record in the Set without adding any new columns)
    - ComputeOperator (executes a computation described in natural language)
    - SearchOperator (executes a search query on the input Context)

    Every logical operator must declare the get_logical_id_params() and get_logical_op_params() methods,
    which return dictionaries of parameters that are used to compute the logical op id and to implement
    the logical operator (respectively).
    """

    def __init__(
        self,
        output_schema: type[BaseModel],
        input_schema: type[BaseModel] | None = None,
        depends_on: list[str] | None = None,
    ):
        # TODO: can we eliminate input_schema?
        self.output_schema = output_schema
        self.input_schema = input_schema
        self.depends_on = [] if depends_on is None else sorted(depends_on)
        self.logical_op_id: str | None = None
        self.unique_logical_op_id: str | None = None

        # compute the fields generated by this logical operator
        input_field_names = list(self.input_schema.model_fields) if self.input_schema is not None else []
        self.generated_fields = sorted(
            [field_name for field_name in self.output_schema.model_fields if field_name not in input_field_names]
        )

    def __str__(self) -> str:
        raise NotImplementedError("Abstract method")

    def __eq__(self, other) -> bool:
        all_id_params_match = all(value == getattr(other, key) for key, value in self.get_logical_id_params().items())
        return isinstance(other, self.__class__) and all_id_params_match

    def copy(self) -> LogicalOperator:
        logical_op_copy = self.__class__(**self.get_logical_op_params())
        logical_op_copy.logical_op_id = self.logical_op_id
        logical_op_copy.unique_logical_op_id = self.unique_logical_op_id
        return logical_op_copy

    def logical_op_name(self) -> str:
        """Name of the logical operator."""
        return str(self.__class__.__name__)

    def get_unique_logical_op_id(self) -> str:
        """
        Get the unique logical operator id for this logical operator.
        """
        return self.unique_logical_op_id

    def set_unique_logical_op_id(self, unique_logical_op_id: str) -> None:
        """
        Set the unique logical operator id for this logical operator.
        This is used to uniquely identify the logical operator in the query plan.
        """
        self.unique_logical_op_id = unique_logical_op_id

    def get_logical_id_params(self) -> dict:
        """
        Returns a dictionary mapping of logical operator parameters which are relevant
        for computing the logical operator id.

        NOTE: Should be overriden by subclasses to include class-specific parameters.
        NOTE: input_schema and output_schema are not included in the id params because
              they depend on how the Optimizer orders operations.
        """
        # TODO: should we use `generated_fields` after getting rid of them in PhysicalOperator?
        return {"generated_fields": self.generated_fields}

    def get_logical_op_params(self) -> dict:
        """
        Returns a dictionary mapping of logical operator parameters which may be used to
        implement a physical operator associated with this logical operation.

        NOTE: Should be overriden by subclasses to include class-specific parameters.
        """
        return {
            "input_schema": self.input_schema,
            "output_schema": self.output_schema,
            "depends_on": self.depends_on,
        }

    def get_logical_op_id(self):
        """
        TODO: turn this into a property?

        NOTE: We do not call this in the __init__() method as subclasses may set parameters
              returned by self.get_logical_op_params() after they call to super().__init__().
        """
        # return self.logical_op_id if we've computed it before
        if self.logical_op_id is not None:
            return self.logical_op_id

        # get op name and op parameters which are relevant for computing the id
        logical_op_name = self.logical_op_name()
        logical_id_params = self.get_logical_id_params()
        logical_id_params = {k: str(v) for k, v in logical_id_params.items()}

        # compute, set, and return the op_id
        hash_str = json.dumps({"logical_op_name": logical_op_name, **logical_id_params}, sort_keys=True)
        self.logical_op_id = hash_for_id(hash_str)

        return self.logical_op_id

    def get_generated_fields(self) -> list[str]:
        """Returns the names of the fields generated by this logical operator."""
        return self.generated_fields

    def __hash__(self):
        if not self.logical_op_id:
            raise ValueError("logical_op_id not set, unable to hash")
        return int(self.logical_op_id, 16)


class Aggregate(LogicalOperator):
    """
    Aggregate is a logical operator that applies an aggregation to the input set and yields a single result.
    This is a base class that has to be further specialized to implement specific aggregation functions.
    """

    def __init__(
        self,
        agg_func: AggFunc | None = None,
        agg_str: str | None = None,
        *args,
        **kwargs,
    ):
        assert agg_func is not None or agg_str is not None, "Either agg_func or agg_str must be provided"
        if kwargs.get("output_schema") is None:
            if agg_func == AggFunc.COUNT:
                kwargs["output_schema"] = Count
            elif agg_func == AggFunc.AVERAGE:
                kwargs["output_schema"] = Average
            elif agg_func == AggFunc.SUM:
                kwargs["output_schema"] = Sum
            elif agg_func == AggFunc.MIN:
                kwargs["output_schema"] = Min
            elif agg_func == AggFunc.MAX:
                kwargs["output_schema"] = Max
            else:
                raise ValueError(f"Unsupported aggregation function: {agg_func}")

        super().__init__(*args, **kwargs)
        self.agg_func = agg_func
        self.agg_str = agg_str

    def __str__(self):
        desc = f"function: {str(self.agg_func.value)}" if self.agg_func else f"agg: {self.agg_str}"
        return f"{self.__class__.__name__}({desc})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "agg_func": self.agg_func,
            "agg_str": self.agg_str,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "agg_func": self.agg_func,
            "agg_str": self.agg_str,
            **logical_op_params,
        }

        return logical_op_params


class BaseScan(LogicalOperator):
    """A BaseScan is a logical operator that represents a scan of a particular root Dataset."""

    def __init__(self, datasource: dataset.Dataset, output_schema: type[BaseModel], *args, **kwargs):
        super().__init__(*args, output_schema=output_schema, **kwargs)
        self.datasource = datasource

    def __str__(self):
        return f"BaseScan({self.datasource},{self.output_schema})"

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, BaseScan)
            and self.input_schema == other.input_schema
            and self.output_schema == other.output_schema
            and self.datasource == other.datasource
        )

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "id": self.datasource.id,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {"datasource": self.datasource, **logical_op_params}

        return logical_op_params


class ContextScan(LogicalOperator):
    """A ContextScan is a logical operator that loads the context for a particular root Dataset."""

    def __init__(self, context: context.Context, output_schema: type[BaseModel], *args, **kwargs):
        super().__init__(*args, output_schema=output_schema, **kwargs)
        self.context = context

    def __str__(self):
        return f"ContextScan({self.context},{self.output_schema})"

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, ContextScan)
            and self.context.id == other.context.id
        )

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "id": self.context.id,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {"context": self.context, **logical_op_params}

        return logical_op_params


class ConvertScan(LogicalOperator):
    """A ConvertScan is a logical operator that represents a scan of a particular input Dataset, with conversion applied."""

    def __init__(
        self,
        cardinality: Cardinality = Cardinality.ONE_TO_ONE,
        udf: Callable | None = None,
        desc: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.cardinality = cardinality
        self.udf = udf
        self.desc = desc

    def __str__(self):
        return f"ConvertScan({self.input_schema} -> {str(self.output_schema)})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "cardinality": self.cardinality,
            "udf": self.udf,
            "desc": self.desc,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "cardinality": self.cardinality,
            "udf": self.udf,
            "desc": self.desc,
            **logical_op_params,
        }

        return logical_op_params


class Distinct(LogicalOperator):
    def __init__(self, distinct_cols: list[str] | None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # if distinct_cols is not None, check that all columns are in the input schema
        if distinct_cols is not None:
            for col in distinct_cols:
                assert col in self.input_schema.model_fields, f"Column {col} not found in input schema {self.input_schema} for Distinct operator"

        # store the list of distinct columns, sorted
        self.distinct_cols = (
            sorted([field_name for field_name in self.input_schema.model_fields])
            if distinct_cols is None
            else sorted(distinct_cols)
        )

    def __str__(self):
        return f"Distinct({self.distinct_cols})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {"distinct_cols": self.distinct_cols, **logical_id_params}

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "distinct_cols": self.distinct_cols,
            **logical_op_params,
        }

        return logical_op_params


class FilteredScan(LogicalOperator):
    """A FilteredScan is a logical operator that represents a scan of a particular input Dataset, with filters applied."""

    def __init__(
        self,
        filter: Filter,
        desc: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.filter = filter
        self.desc = desc

    def __str__(self):
        return f"FilteredScan({str(self.output_schema)}, {str(self.filter)})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "filter": self.filter,
            "desc": self.desc,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "filter": self.filter,
            "desc": self.desc,
            **logical_op_params,
        }

        return logical_op_params


class GroupByAggregate(LogicalOperator):
    def __init__(
        self,
        group_by_sig: GroupBySig,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if not self.input_schema:
            raise ValueError("GroupByAggregate requires an input schema")
        (valid, error) = group_by_sig.validate_schema(self.input_schema)
        if not valid:
            raise TypeError(error)
        self.group_by_sig = group_by_sig

    def __str__(self):
        return f"GroupBy({self.group_by_sig.serialize()})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {"group_by_sig": self.group_by_sig, **logical_id_params}

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "group_by_sig": self.group_by_sig,
            **logical_op_params,
        }

        return logical_op_params


class JoinOp(LogicalOperator):
    def __init__(self, condition: str, on: list[str] | None = None, how: str = "inner", desc: str | None = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.condition = condition
        self.on = on
        self.how = how
        self.desc = desc

    def __str__(self):
        return f"Join(condition={self.condition})" if self.on is None else f"Join(on={self.on}, how={self.how})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "condition": self.condition,
            "on": self.on,
            "how": self.how,
            "desc": self.desc,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "condition": self.condition,
            "on": self.on,
            "how": self.how,
            "desc": self.desc,
            **logical_op_params,
        }

        return logical_op_params


class LimitScan(LogicalOperator):
    def __init__(self, limit: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.limit = limit

    def __str__(self):
        return f"LimitScan({str(self.input_schema)}, {str(self.output_schema)})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {"limit": self.limit, **logical_id_params}

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "limit": self.limit,
            **logical_op_params,
        }

        return logical_op_params


class Project(LogicalOperator):
    def __init__(self, project_cols: list[str], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.project_cols = project_cols

    def __str__(self):
        return f"Project({self.input_schema}, {self.project_cols})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {"project_cols": self.project_cols, **logical_id_params}

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "project_cols": self.project_cols,
            **logical_op_params,
        }

        return logical_op_params


class TopKScan(LogicalOperator):
    """A TopKScan is a logical operator that represents a scan of a particular input Dataset, with a top-k operation applied."""

    def __init__(
        self,
        index,
        search_func,
        search_attr,
        output_attrs,
        k,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.index = index
        self.search_func = search_func
        self.search_attr = search_attr
        self.output_attrs = output_attrs
        self.k = k

    def __str__(self):
        return f"TopKScan({self.input_schema} -> {str(self.output_schema)})"

    def get_logical_id_params(self) -> dict:
        # NOTE: if we allow optimization over index, then we will need to include it in the id params
        # NOTE: ^if we do this, we should probably make a wrapper around the index object to ensure that
        #       it can be serialized as a string properly
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "search_attr": self.search_attr,
            "output_attrs": self.output_attrs,
            "k": self.k,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "index": self.index,
            "search_func": self.search_func,
            "search_attr": self.search_attr,
            "output_attrs": self.output_attrs,
            "k": self.k,
            **logical_op_params,
        }

        return logical_op_params


class ComputeOperator(LogicalOperator):
    """
    A ComputeOperator is a logical operator that performs a computation described in natural language
    on a given Context.
    """

    def __init__(self, context_id: str, instruction: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.context_id = context_id
        self.instruction = instruction

    def __str__(self):
        return f"ComputeOperator(id={self.context_id}, instr={self.instruction:20s})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "context_id": self.context_id,
            "instruction": self.instruction,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "context_id": self.context_id,
            "instruction": self.instruction,
            **logical_op_params,
        }

        return logical_op_params


class SearchOperator(LogicalOperator):
    """
    A SearchOperator is a logical operator that executes a search described in natural language
    on a given Context.
    """

    def __init__(self, context_id: str, search_query: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.context_id = context_id
        self.search_query = search_query

    def __str__(self):
        return f"SearchOperator(id={self.context_id}, search_query={self.search_query:20s})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "context_id": self.context_id,
            "search_query": self.search_query,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "context_id": self.context_id,
            "search_query": self.search_query,
            **logical_op_params,
        }

        return logical_op_params



import functools
import inspect
import os
import time
from typing import Any

# from mem0 import Memory
from smolagents import CodeAgent, LiteLLMModel, tool

# from palimpzest.agents.search_agents import DataDiscoveryAgent, SearchManagerAgent
from palimpzest.core.data.context import Context
from palimpzest.core.data.context_manager import ContextManager
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.models import GenerationStats, OperatorCostEstimates, RecordOpStats
from palimpzest.query.operators.physical import PhysicalOperator


def make_tool(bound_method):
    # Get the original function and bound instance
    func = bound_method.__func__
    instance = bound_method.__self__
    
    # Get the signature and remove 'self'
    sig = inspect.signature(func)
    params = list(sig.parameters.values())[1:]  # skip 'self'
    new_sig = inspect.Signature(parameters=params, return_annotation=sig.return_annotation)

    # Create a wrapper function dynamically
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(instance, *args, **kwargs)

    # Update the __signature__ to reflect the new one without 'self'
    wrapper.__signature__ = new_sig

    return wrapper


class SmolAgentsSearch(PhysicalOperator):
    """
    Physical operator for searching with Smol Agents.
    """
    def __init__(self, context_id: str, search_query: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.context_id = context_id
        self.search_query = search_query
        # self.model_id = "anthropic/claude-3-7-sonnet-latest"
        self.model_id = "openai/gpt-4o-mini-2024-07-18"
        # self.model_id = "openai/gpt-4o-2024-08-06"
        api_key = os.getenv("ANTHROPIC_API_KEY") if "anthropic" in self.model_id else os.getenv("OPENAI_API_KEY")
        self.model = LiteLLMModel(model_id=self.model_id, api_key=api_key)

    def __str__(self):
        op = super().__str__()
        op += f"    Context ID: {self.context_id:20s}\n"
        op += f"    Search Query: {self.search_query:20s}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {
            "context_id": self.context_id,
            "search_query": self.search_query,
            **id_params,
        }

    def get_op_params(self):
        op_params = super().get_op_params()
        return {
            "context_id": self.context_id,
            "search_query": self.search_query,
            **op_params,
        }

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        return OperatorCostEstimates(
            cardinality=source_op_cost_estimates.cardinality,
            time_per_record=100,
            cost_per_record=1,
            quality=1.0,
        )

    def _create_record_set(
        self,
        candidate: DataRecord,
        generation_stats: GenerationStats,
        total_time: float,
        answer: dict[str, Any],
    ) -> DataRecordSet:
        """
        Given an input DataRecord and a determination of whether it passed the filter or not,
        construct the resulting RecordSet.
        """
        # create new DataRecord
        data_item = {field: answer[field] for field in self.output_schema.model_fields if field in answer}
        dr = DataRecord.from_parent(self.output_schema, data_item, parent_record=candidate)

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr._id,
            record_parent_ids=dr._parent_ids,
            record_source_indices=dr._source_indices,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=total_time,
            cost_per_record=generation_stats.cost_per_record,
            model_name=self.get_model_name(),
            total_input_tokens=generation_stats.total_input_tokens,
            total_output_tokens=generation_stats.total_output_tokens,
            total_input_cost=generation_stats.total_input_cost,
            total_output_cost=generation_stats.total_output_cost,
            llm_call_duration_secs=generation_stats.llm_call_duration_secs,
            fn_call_duration_secs=generation_stats.fn_call_duration_secs,
            total_llm_calls=generation_stats.total_llm_calls,
            total_embedding_llm_calls=generation_stats.total_embedding_llm_calls,
            answer={k: v.description if isinstance(v, Context) else v for k, v in answer.items()},
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return DataRecordSet([dr], [record_op_stats])

    def __call__(self, candidate: DataRecord) -> Any:
        start_time = time.time()

        # get the input context object and its tools
        input_context: Context = candidate.context
        description = input_context.description
        tools = [tool(make_tool(f)) for f in input_context.tools]

        # # construct the full search query
        # full_query = f"Please execute the following search query. Output a **detailed** description of (1) which data you look at, and (2) what you find in that data. Avoid making overly broad statements such as \"What you're searching for is not present in the dataset\". Instead, make more precise statments like \"What you're searching for is not present in files A.txt, B.txt, and C.txt, but may be present elsewhere\".\n\nQUERY: {self.search_query}"

        # perform the computation
        instructions = f"\n\nHere is a description of the Context whose data you will be working with, as well as any previously computed results:\n\n{description}"
        agent = CodeAgent(
            tools=tools,
            model=self.model,
            add_base_tools=False,
            instructions=instructions,
            return_full_result=True,
            additional_authorized_imports=["pandas", "io", "os"],
        )
        result = agent.run(self.search_query)
        # NOTE: you can see the system prompt with `agent.memory.system_prompt.system_prompt`
        # full_steps = agent.memory.get_full_steps()

        # compute generation stats
        response = result.output
        input_tokens = result.token_usage.input_tokens
        output_tokens = result.token_usage.output_tokens
        cost_per_input_token = (3.0 / 1e6) if "anthropic" in self.model_id else (0.15 / 1e6) # (2.5 / 1e6) #
        cost_per_output_token = (15.0 / 1e6) if "anthropic" in self.model_id else (0.6 / 1e6) # (10.0 / 1e6) #
        input_cost = input_tokens * cost_per_input_token
        output_cost = output_tokens * cost_per_output_token
        generation_stats = GenerationStats(
            model_name=self.model_id,
            total_input_tokens=input_tokens,
            total_output_tokens=output_tokens,
            total_input_cost=input_cost,
            total_output_cost=output_cost,
            cost_per_record=input_cost + output_cost,
            llm_call_duration_secs=time.time() - start_time,
        )

        # update the description of the Context to include the search result
        new_description = f"RESULT: {response}\n\n"
        cm = ContextManager()
        cm.update_context(id=self.context_id, description=new_description)

        # create and return record set
        field_answers = {
            "context": cm.get_context(id=self.context_id),
        }
        record_set = self._create_record_set(
            candidate,
            generation_stats,
            time.time() - start_time,
            field_answers,
        )

        return record_set


# class SmolAgentsManagedSearch(PhysicalOperator):
#     """
#     Physical operator for searching with Smol Agents using an Orchestrator and a Data Discovery Agent.
#     """
#     def __init__(self, context_id: str, search_query: str, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.context_id = context_id
#         self.search_query = search_query
#         # self.model_id = "anthropic/claude-3-7-sonnet-latest"
#         self.model_id = "openai/gpt-4o-mini-2024-07-18"
#         # self.model_id = "o1"
#         model_params = {
#             "model_id": self.model_id,
#             "custom_role_conversions": {"tool-call": "assistant", "tool-response": "user"},
#             "max_completion_tokens": 8192,
#         }
#         if self.model_id == "o1":
#             model_params["reasoning_effort"] = "high"
#         self.model = LiteLLMModel(**model_params)
#         self.text_limit = 100000
#         self.memory = Memory()

#     def __str__(self):
#         op = super().__str__()
#         op += f"    Context ID: {self.context_id:20s}\n"
#         op += f"    Search Query: {self.search_query:20s}\n"
#         return op

#     def get_id_params(self):
#         id_params = super().get_id_params()
#         return {
#             "context_id": self.context_id,
#             "search_query": self.search_query,
#             **id_params,
#         }

#     def get_op_params(self):
#         op_params = super().get_op_params()
#         return {
#             "context_id": self.context_id,
#             "search_query": self.search_query,
#             **op_params,
#         }

#     def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
#         return OperatorCostEstimates(
#             cardinality=source_op_cost_estimates.cardinality,
#             time_per_record=100,
#             cost_per_record=1,
#             quality=1.0,
#         )

#     def _create_record_set(
#         self,
#         candidate: DataRecord,
#         generation_stats: GenerationStats,
#         total_time: float,
#         answer: dict[str, Any],
#     ) -> DataRecordSet:
#         """
#         Given an input DataRecord and a determination of whether it passed the filter or not,
#         construct the resulting RecordSet.
#         """
#         # create new DataRecord
#         data_item = {field: answer[field] for field in self.output_schema.model_fields if field in answer}
#         dr = DataRecord.from_parent(self.output_schema, data_item, parent_record=candidate)

        # # create RecordOpStats object
        # record_op_stats = RecordOpStats(
        #     record_id=dr._id,
        #     record_parent_ids=dr._parent_ids,
        #     record_source_indices=dr._source_indices,
        #     record_state=dr.to_dict(include_bytes=False),
        #     full_op_id=self.get_full_op_id(),
        #     logical_op_id=self.logical_op_id,
        #     op_name=self.op_name(),
        #     time_per_record=total_time,
        #     cost_per_record=generation_stats.cost_per_record,
        #     model_name=self.get_model_name(),
        #     total_input_tokens=generation_stats.total_input_tokens,
        #     total_output_tokens=generation_stats.total_output_tokens,
        #     total_input_cost=generation_stats.total_input_cost,
        #     total_output_cost=generation_stats.total_output_cost,
        #     llm_call_duration_secs=generation_stats.llm_call_duration_secs,
        #     fn_call_duration_secs=generation_stats.fn_call_duration_secs,
        #     total_llm_calls=generation_stats.total_llm_calls,
        #     total_embedding_llm_calls=generation_stats.total_embedding_llm_calls,
        #     answer={k: v.description if isinstance(v, Context) else v for k, v in answer.items()},
        #     op_details={k: str(v) for k, v in self.get_id_params().items()},
        # )

#         return DataRecordSet([dr], [record_op_stats])

#     def __call__(self, candidate: DataRecord) -> Any:
#         start_time = time.time()

#         # get the input context object and its tools
#         input_context: Context = candidate.context
#         description = input_context.description
#         tools = [tool(make_tool(f)) for f in input_context.tools]

#         # create a memory tool for accessing past searches
#         @tool
#         def tool_search_history(query: str) -> str:
#             """
#             This tool enables the agent to search through its history of execution in previous sessions.
#             Thus, the agent can learn more about what it has done in the past by invoking this tool with
#             a query describing what past interactions the agent might be curious about.

#             Args:
#                 query (str): A description of what the agent wishes to search for in its execution history.

#             Returns:
#                 str: A summary of the agent execution history which is relevant to the query.
#             """
#             memories = self.memory.search(query=query, user_id="data_discovery_agent")
#             memory_str = ""
#             for idx, memory in enumerate(memories):
#                 memory_str += f"MEMORY {idx+1}: {memory['memory']}"
#             return memory_str

#         # tools.append(tool_search_history)
#         data_discovery_agent = CodeAgent(
#             model=self.model,
#             tools=tools,
#             max_steps=20,
#             verbosity_level=2,
#             planning_interval=4,
#             name="data_discovery_agent",
#             description="""A team member that will search a data repository to find files which help to answer your question.
#         Ask him for all your questions that require searching a repository of relevant data.
#         Provide him as much context as possible, in particular if you need to search on a specific timeframe!
#         And don't hesitate to provide him with a complex search task, like finding a difference between two files.
#         Your request must be a real sentence, not a keyword search! Like "Find me this information (...)" rather than a few keywords.
#         """,
#             provide_run_summary=True,
#         )
#         data_discovery_agent.prompt_templates["managed_agent"]["task"] += f"""\n\nHere is a description of the context you will be working with: {description}\n\nSearch as many files as possible before returning your final answer.\n\nAdditionally, if after some searching you find out that you need more information to answer the question, you can use `final_answer` with your request for clarification as argument to request for more information."""

#         manager_agent = CodeAgent(
#             model=self.model,
#             tools=tools,
#             max_steps=12,
#             verbosity_level=2,
#             additional_authorized_imports=["*"],
#             planning_interval=4,
#             managed_agents=[data_discovery_agent],
#             return_full_result=True,
#         )

#         # TODO: improve context descriptions and add memory from there; expand to multi-modal benchmark(s)
#         # perform the computation
#         result = manager_agent.run(self.search_query)

#         # compute generation stats
#         response = result.output
#         input_tokens = result.token_usage.input_tokens
#         output_tokens = result.token_usage.output_tokens
#         cost_per_input_token = (3.0 / 1e6) if "anthropic" in self.model_id else (0.15 / 1e6) # (15.0 / 1e6)
#         cost_per_output_token = (15.0 / 1e6) if "anthropic" in self.model_id else (0.6 / 1e6) # (60.0 / 1e6)
#         input_cost = input_tokens * cost_per_input_token
#         output_cost = output_tokens * cost_per_output_token
#         generation_stats = GenerationStats(
#             model_name=self.model_id,
#             total_input_tokens=input_tokens,
#             total_output_tokens=output_tokens,
#             total_input_cost=input_cost,
#             total_output_cost=output_cost,
#             cost_per_record=input_cost + output_cost,
#             llm_call_duration_secs=time.time() - start_time,
#         )

#         # update the description of the Context to include the search result
#         new_description = f"RESULT: {response}\n\n"
#         cm = ContextManager()
#         cm.update_context(id=self.context_id, description=new_description)

#         # create and return record set
#         field_answers = {
#             "context": cm.get_context(id=self.context_id),
#         }
#         record_set = self._create_record_set(
#             candidate,
#             generation_stats,
#             time.time() - start_time,
#             field_answers,
#         )

#         return record_set


# class SmolAgentsCustomManagedSearch(PhysicalOperator):
#     """
#     Physical operator for searching with Smol Agents using an Orchestrator and a Data Discovery Agent.
#     """
#     def __init__(self, context_id: str, search_query: str, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.context_id = context_id
#         self.search_query = search_query
#         # self.model_id = "anthropic/claude-3-7-sonnet-latest"
#         self.model_id = "openai/gpt-4o-mini-2024-07-18"
#         # self.model_id = "o1"
#         model_params = {
#             "model_id": self.model_id,
#             "custom_role_conversions": {"tool-call": "assistant", "tool-response": "user"},
#             "max_completion_tokens": 8192,
#         }
#         if self.model_id == "o1":
#             model_params["reasoning_effort"] = "high"
#         self.model = LiteLLMModel(**model_params)
#         self.text_limit = 100000

#     def __str__(self):
#         op = super().__str__()
#         op += f"    Context ID: {self.context_id:20s}\n"
#         op += f"    Search Query: {self.search_query:20s}\n"
#         return op

#     def get_id_params(self):
#         id_params = super().get_id_params()
#         return {
#             "context_id": self.context_id,
#             "search_query": self.search_query,
#             **id_params,
#         }

#     def get_op_params(self):
#         op_params = super().get_op_params()
#         return {
#             "context_id": self.context_id,
#             "search_query": self.search_query,
#             **op_params,
#         }

#     def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
#         return OperatorCostEstimates(
#             cardinality=source_op_cost_estimates.cardinality,
#             time_per_record=100,
#             cost_per_record=1,
#             quality=1.0,
#         )

#     def _create_record_set(
#         self,
#         candidate: DataRecord,
#         generation_stats: GenerationStats,
#         total_time: float,
#         answer: dict[str, Any],
#     ) -> DataRecordSet:
#         """
#         Given an input DataRecord and a determination of whether it passed the filter or not,
#         construct the resulting RecordSet.
#         """
#         # create new DataRecord
#         data_item = {field: answer[field] for field in self.output_schema.model_fields if field in answer}
#         dr = DataRecord.from_parent(self.output_schema, data_item, parent_record=candidate)

#         # create RecordOpStats object
#         record_op_stats = RecordOpStats(
#             record_id=dr._id,
#             record_parent_ids=dr._parent_ids,
#             record_source_indices=dr._source_indices,
#             record_state=dr.to_dict(include_bytes=False),
#             full_op_id=self.get_full_op_id(),
#             logical_op_id=self.logical_op_id,
#             op_name=self.op_name(),
#             time_per_record=total_time,
#             cost_per_record=generation_stats.cost_per_record,
#             model_name=self.get_model_name(),
#             total_input_tokens=generation_stats.total_input_tokens,
#             total_output_tokens=generation_stats.total_output_tokens,
#             total_input_cost=generation_stats.total_input_cost,
#             total_output_cost=generation_stats.total_output_cost,
#             llm_call_duration_secs=generation_stats.llm_call_duration_secs,
#             fn_call_duration_secs=generation_stats.fn_call_duration_secs,
#             total_llm_calls=generation_stats.total_llm_calls,
#             total_embedding_llm_calls=generation_stats.total_embedding_llm_calls,
#             answer={k: v.description if isinstance(v, Context) else v for k, v in answer.items()},
#             op_details={k: str(v) for k, v in self.get_id_params().items()},
#         )

#         return DataRecordSet([dr], [record_op_stats])

#     def __call__(self, candidate: DataRecord) -> Any:
#         start_time = time.time()

#         # get the input context object and its tools
#         input_context: Context = candidate.context
#         description = input_context.description
#         tools = [tool(make_tool(f)) for f in input_context.tools]

#         # TODO: add semantic operators to tools
#         data_discovery_agent = DataDiscoveryAgent(self.context_id, description, model=self.model, tools=tools)
#         search_manager_agent = SearchManagerAgent(self.context_id, description, model=self.model, tools=tools, managed_agents=[data_discovery_agent])

#         # perform the computation
#         result = search_manager_agent.run(self.search_query)

#         # compute generation stats
#         response = result.output
#         input_tokens = result.token_usage.input_tokens
#         output_tokens = result.token_usage.output_tokens
#         cost_per_input_token = (3.0 / 1e6) if "anthropic" in self.model_id else (0.15 / 1e6) # (15.0 / 1e6)
#         cost_per_output_token = (15.0 / 1e6) if "anthropic" in self.model_id else (0.6 / 1e6) # (60.0 / 1e6)
#         input_cost = input_tokens * cost_per_input_token
#         output_cost = output_tokens * cost_per_output_token
#         generation_stats = GenerationStats(
#             model_name=self.model_id,
#             total_input_tokens=input_tokens,
#             total_output_tokens=output_tokens,
#             total_input_cost=input_cost,
#             total_output_cost=output_cost,
#             cost_per_record=input_cost + output_cost,
#             llm_call_duration_secs=time.time() - start_time,
#         )

#         # update the description of the Context to include the search result
#         new_description = f"RESULT: {response}\n\n"
#         cm = ContextManager()
#         cm.update_context(id=self.context_id, description=new_description)

#         # create and return record set
#         field_answers = {
#             "context": cm.get_context(id=self.context_id),
#         }
#         record_set = self._create_record_set(
#             candidate,
#             generation_stats,
#             time.time() - start_time,
#             field_answers,
#         )

#         return record_set



from palimpzest.query.operators.aggregate import AggregateOp as _AggregateOp
from palimpzest.query.operators.aggregate import ApplyGroupByOp as _ApplyGroupByOp
from palimpzest.query.operators.aggregate import AverageAggregateOp as _AverageAggregateOp
from palimpzest.query.operators.aggregate import CountAggregateOp as _CountAggregateOp
from palimpzest.query.operators.aggregate import MaxAggregateOp as _MaxAggregateOp
from palimpzest.query.operators.aggregate import MinAggregateOp as _MinAggregateOp
from palimpzest.query.operators.aggregate import SemanticAggregate as _SemanticAggregate
from palimpzest.query.operators.aggregate import SumAggregateOp as _SumAggregateOp
from palimpzest.query.operators.convert import ConvertOp as _ConvertOp
from palimpzest.query.operators.convert import LLMConvert as _LLMConvert
from palimpzest.query.operators.convert import LLMConvertBonded as _LLMConvertBonded
from palimpzest.query.operators.convert import NonLLMConvert as _NonLLMConvert
from palimpzest.query.operators.critique_and_refine import CritiqueAndRefineConvert as _CritiqueAndRefineConvert
from palimpzest.query.operators.critique_and_refine import CritiqueAndRefineFilter as _CritiqueAndRefineFilter
from palimpzest.query.operators.distinct import DistinctOp as _DistinctOp
from palimpzest.query.operators.filter import FilterOp as _FilterOp
from palimpzest.query.operators.filter import LLMFilter as _LLMFilter
from palimpzest.query.operators.filter import NonLLMFilter as _NonLLMFilter
from palimpzest.query.operators.join import EmbeddingJoin as _EmbeddingJoin
from palimpzest.query.operators.join import JoinOp as _JoinOp
from palimpzest.query.operators.join import NestedLoopsJoin as _NestedLoopsJoin
from palimpzest.query.operators.limit import LimitScanOp as _LimitScanOp
from palimpzest.query.operators.logical import (
    Aggregate as _Aggregate,
)
from palimpzest.query.operators.logical import (
    BaseScan as _BaseScan,
)
from palimpzest.query.operators.logical import (
    ConvertScan as _ConvertScan,
)
from palimpzest.query.operators.logical import (
    Distinct as _Distinct,
)
from palimpzest.query.operators.logical import (
    FilteredScan as _FilteredScan,
)
from palimpzest.query.operators.logical import (
    GroupByAggregate as _GroupByAggregate,
)
from palimpzest.query.operators.logical import (
    JoinOp as _LogicalJoinOp,
)
from palimpzest.query.operators.logical import (
    LimitScan as _LimitScan,
)
from palimpzest.query.operators.logical import (
    LogicalOperator as _LogicalOperator,
)
from palimpzest.query.operators.logical import (
    Project as _Project,
)
from palimpzest.query.operators.logical import (
    TopKScan as _TopKScan,
)
from palimpzest.query.operators.mixture_of_agents import MixtureOfAgentsConvert as _MixtureOfAgentsConvert
from palimpzest.query.operators.mixture_of_agents import MixtureOfAgentsFilter as _MixtureOfAgentsFilter
from palimpzest.query.operators.physical import PhysicalOperator as _PhysicalOperator
from palimpzest.query.operators.project import ProjectOp as _ProjectOp
from palimpzest.query.operators.rag import RAGConvert as _RAGConvert
from palimpzest.query.operators.rag import RAGFilter as _RAGFilter
from palimpzest.query.operators.scan import MarshalAndScanDataOp as _MarshalAndScanDataOp
from palimpzest.query.operators.scan import ScanPhysicalOp as _ScanPhysicalOp
from palimpzest.query.operators.split import SplitConvert as _SplitConvert
from palimpzest.query.operators.split import SplitFilter as _SplitFilter
from palimpzest.query.operators.topk import TopKOp as _TopKOp

LOGICAL_OPERATORS = [
    _LogicalOperator,
    _Aggregate,
    _BaseScan,
    _ConvertScan,
    _Distinct,
    _FilteredScan,
    _GroupByAggregate,
    _LogicalJoinOp,
    _LimitScan,
    _Project,
    _TopKScan,
]

PHYSICAL_OPERATORS = (
    # aggregate
    [_AggregateOp, _ApplyGroupByOp, _AverageAggregateOp, _CountAggregateOp, _MaxAggregateOp, _MinAggregateOp, _SemanticAggregate, _SumAggregateOp]
    # convert
    + [_ConvertOp, _NonLLMConvert, _LLMConvert, _LLMConvertBonded]
    # critique and refine
    + [_CritiqueAndRefineConvert, _CritiqueAndRefineFilter]
    # distinct
    + [_DistinctOp]
    # scan
    + [_ScanPhysicalOp, _MarshalAndScanDataOp]
    # filter
    + [_FilterOp, _NonLLMFilter, _LLMFilter]
    # join
    + [_EmbeddingJoin, _JoinOp, _NestedLoopsJoin]
    # limit
    + [_LimitScanOp]
    # mixture-of-agents
    + [_MixtureOfAgentsConvert, _MixtureOfAgentsFilter]
    # physical
    + [_PhysicalOperator]
    # project
    + [_ProjectOp]
    # rag
    + [_RAGConvert, _RAGFilter]
    # top-k
    + [_TopKOp]
    # split
    + [_SplitConvert, _SplitFilter]
)

__all__ = [
    "LOGICAL_OPERATORS",
    "PHYSICAL_OPERATORS",
]
